require 'base.scripts.menu'

------------------ remote menu system ------------------

	-- serverside

local menuServerSide = {}
menuServerSide.remoteMenuID = 1	-- stored in the class, incremented serverside

function menuServerSide:getAndIncrementRemoteMenuID()
	local id = tostring(self.class.remoteMenuID)
	self.class.remoteMenuID = self.class.remoteMenuID + 1
	return id
end

function menuServerSide:createPromptingMenu(args)
	-- [[lock me please
	local oldControl = self.control
	self.control = false

	-- [[lock me please
	-- note that we only need a map if multiple parallel prompt/server menus can be open at the same time
	-- right now that's not the case
	self.remoteMenuActive = self.remoteMenuActive or {}
	local id = self:getAndIncrementRemoteMenuID()
	self.remoteMenuActive[id] = true
	--]]
	
	local remoteargs = table.new()
	for k,v in pairs(args) do
		remoteargs:insert(k)
		remoteargs:insert(v)
	end
	self:serverSendMessage('clientMenuCreate', id, unpack(remoteargs))
	
	while true do
		-- [[lock me please
		if not self.remoteMenuActive[id] then break end
		--]]
		sleep(.05)
	end

	self.control = oldControl
	--]]
end

function menuServerSide:serverRemoteMenuClosed(id)
	self.remoteMenuActive[id] = nil
end


	-- clientside

-- menus that the server has requested we create
-- it gives unique ids too, so store them by those ids
ClientSide.serverMenus = {}

function ServerMenuOnDelete(self)
	ClientSide.serverMenus[self.serverID] = nil
	entClientSendMessage(self.client.entID, 'serverRemoteMenuClosed', self.serverID)
end

function ClientSide:clientMenuCreate(id, ...)
	local args = {...}
	-- fill in default keys
	local menuclass = Menu
	local menuargs = {
		client=self;
		serverID=id;
		onDelete=ServerMenuOnDelete;	-- tell the server it's gone
	}
	local i = 1
	-- calc the rest
	-- note, for menu handles, esp stating parents, we'll have to start thinking
	while i <= #args do
		assert(i < #args, "got a mismatched number of keys and values")
		local key = args[i]
		i=i+1
		local value = args[i]
		i=i+1
		-- executing remotely sent code, eh? TODO - security?!
		value = exec('return '..value)
		
		if key == 'class' then
			menuclass = value
		else
			menuargs[key] = value
		end
	end
	self.serverMenus[id] = menuclass:new(menuargs)
end

function ClientSide:clientMenuClose(id)
	local m = self.serverMenus[id]
	if m and m.active then m:delete() end
end

------------------ dialog menus ------------------

	-- serverside


function menuServerSide:doDialog(args)
	self:createPromptingMenu{
		class='DialogMenu';
		text=string.format('%q', assert(args.text));
		duration=args.duration;
	}
end

function menuServerSide:dialogDuration(duration,text)
	self:doDialog{duration=duration,text=text}
end

function menuServerSide:dialog(d)
	-- this is only for prompting dialogs.  there is such a thing as promptless...	
	if type(d) == 'string' then
		self:doDialog{text=d}
	elseif type(d) == 'table' then
		for _,v in ipairs(d) do
			self:doDialog{text=v}
		end
	end
end

	-- clientside

class('DialogMenu', Menu)
DialogMenu.pos = {1,1}
DialogMenu.size = {Menu.sysSize[1]-2, 15}
DialogMenu.hasPrompts = true

function DialogMenu:ctor(...)
	DialogMenu.super.ctor(self, ...)
	local text = self.text
	self.text = nil
	local ch = Menu:new{
		parent=self;
		pos={1,1};
		size=self.size - {2,2};
		text=text;
		backgroundColor={0,0,0,0};
		isPrompt=true;
		onOk = function(self)
			self.parent:delete()
		end;
	}
	
	self.startTime = getTime()	--clientside as well, eh?
end

function DialogMenu:onOk(input)
	self:delete() 	-- deleting the menu mid input handling will mess up the iterator ... gracefully?
end

function DialogMenu:update(...)
	if self.duration and getTime() - self.startTime > self.duration then
		self:delete()
		return
	end
	return DialogMenu.super.update(self, ...)
end


------------------ exposure to whoever wants it ------------------
-- preferrably, if not particularly, the client 


function makeEntityMenuCapable(cl)
	union(cl, menuServerSide)
end
