-- utility

function readVec(s)
	local index = 1
	local v = {}
	local si = 1
	while si <= string.len(s) do
		local next = string.find(s, ' ', si, true)
		v[index] = tonumber(string.sub(s, si, next))
		index = index + 1
		if not next then break end
		si = next + 1
	end
	return v
end

function readVecN(s,n)
	local v = readVec(s)
	for i = 1,n do
		if not v[i] then v[i] = 0 end
	end
	return v
end

table.__index = table
function table.new(...)		-- union by default.  makes "union" obsolete.  but subtly dif in behavior (1st param isnt returned, instead a new table is)
	local t = {}
	setmetatable(t, table)
	for _,o in ipairs{...} do
		for k,v in pairs(o) do
			t[k] = v
		end
	end
	return t
end

-- not in luaext
-- cycles though all args past the first and sets all entries in them into the first
function union(t,...)
	for _,o in ipairs{...} do
		for k,v in pairs(o) do
			t[k] = v
		end
	end
	return t
end
-- might as well make this easy on tables
table.union = union

-- not in luaext
-- same as above, but only for entries not in ta
-- actually this is the same as above if you reverse the order -- and pass a first arg of {}
function unionlite(ta,tb)
	for k,v in pairs(tb) do
		if ta[k] == nil then ta[k] = v end
	end
	return ta
end
table.unionLite = unionlite

function table:append(...)
	-- should we have two nested loops like this or should we force users to unpack() all the passed arrays?
	for _,u in ipairs{...} do
		for _,v in ipairs(u) do
			table.insert(self, v)
		end
	end
	return t
end

-- not in luaext
function table.prepend(t, ...)
	-- should we have two nested loops like this or should we force users to unpack() all the passed arrays?
	for _,o in ipairs{...} do
		for _,v in ipairs(o) do
			table.insert(t, 1, v)
		end
	end
	return t
end

function table:find(value, eq)
	if eq then
		for k,v in pairs(self) do
			if eq(v, value) then return k end
		end
	else
		for k,v in pairs(self) do
			if v == value then return k end
		end
	end
end

function table.insertUnique(t, obj)
	if not table.find(t, obj) then table.insert(t, obj) end
end

-- not in luaext
function table.removeObject(t, obj)
	local k = table.find(t, obj)
	while k ~= nil do
		if type(k) == 'number' then
			table.remove(t, k)
		else
			t[k] = nil
		end
		k = table.find(t, obj)
	end
end

function table:map(cb)
	local t = table.new()
	for k,v in pairs(self) do
		local nk,nv = cb(k,v)
		t[nk] = nv
	end
	return t
end

-- accepts a result table to be appended and an input array mapping naturals to key strings
function enum(keys, initValue)
	if not initValue then initValue = 1 end
	local t = {}
	for i,v in ipairs(keys) do
		t[v] = initValue + i - 1	-- so initValue is the first value in the enum
	end
	return t
end

-- appends the string to every element in the set
function setAppendStringPrefix(set, str)
	for k,v in pairs(set) do
		set[k] = str .. tostring(v)
	end
	return set
end

function istable(b)
	return type(b)=='table'
end

function isnumber(b)
	return type(b)=='number'
end

function castnumber(b)
	return tonumber(b) or 0
end

-- is this hardcoded?
-- float (float,float,float)
function clamp(a,min,max)
	if a < min then
		return min
	elseif a > max then
		return max
	end
	return a
end

-- note that while loadstring('2')() won't return 2, loadstring('return 2')() will return 2
function exec(str, chunk)
	local fn, errmsg
	fn, errmsg = loadstring(str, chunk)
	if not fn then
		print('ERROR:'..tostring(errmsg))
		print('CHUNK:'..chunk)
		print('CODE:\n'..str)
		assert(false, 'ERROR - loadstring ' .. chunk .. ' failed!')
	end
	return fn()
end

-- run multiple things in parallel using threads
function parallel(...)
	local threads = table.new()
	for i,func in ipairs{...} do
		threads:insert(fork(func))
	end
	
	for i,th in ipairs(threads) do
		wait(th)
	end
end

function timedRepeat(callback, duration)
	local startTime = getTime()
	repeat
		local soFar = getTime() - startTime
		-- this is a constraint common enough i'm just going to put it here until I find even one example that doesn't require it
		if soFar > duration then soFar = duration end
		callback(soFar)
		sleep(0)
	until soFar >= duration
end

-- string

function string.split(s, exp)
	s = tostring(s)
	local t = table.new()
	local searchpos = 1
	local start, fin = s:find(exp, searchpos)
	while start do
		t:insert(s:sub(searchpos, start-1))
		searchpos = fin+1
		start, fin = s:find(exp, searchpos)
	end
	t:insert(s:sub(searchpos))
	return t
end

function string.trim(s)
	return s:match('^%s*(.-)%s*$')
end

function string.bytes(s)
	return table.new{s:byte(1,#s)}
end

-- io

function io.readfile(fn)
	local f = io.open(fn, 'r')
	if not f then return false end
	local d = f:read('*a')
	f:close()
	return d
end

function io.writefile(fn, d)
	local f = io.open(fn, 'w')
	if not f then return false end
	f:write(d)
	f:close()
	return true
end

function io.fileexists(fn)
	local f = io.open(fn, 'r')
	if not f then return false end
	f:close()
	return true
end

