require 'bomberman.scripts.base'
require 'base.scripts.shader'	--BombWarp

animLoad('bomb', 'bomb.png')

entclass('Bomb', Base)

Bomb.animGroup = 'bomb'
Bomb.bbox = box2:new(.45)	--block players going down single tiled aisles
Bomb.speed = 10

Bomb.duration = 2
Bomb.chainTime = .1

function Bomb:ctor(owner, bombpos, bombLength)
	Bomb.super.ctor(self)

	owner.bombs:insert(self)		-- keep track of all bombs an owner owns for their shot's sake

	--snap to closest half tile
	self:setPos(bombpos)

	self.boomTime = getTime() + self.duration
	self.bombLength = bombLength
	self.owner = owner

	-- net hack vars (for less ent alloc and hopefully faster render time)
	self.boomDists = vec4:new()
end

-- make a super for all bomberman classes that are bombable and just put this in there
function Bomb:delete()
	--remove unique form owner.bombs
	for i,v in ipairs(self.owner.bombs) do
		if v == self then
			self.owner.bombs:remove(i)	--remove operates by index
			break		-- assume it's unique and stop here (otherwise we'd have to mess with the (readonly) iterator)
		end
	end	

	Bomb.super.delete(self)
end

function Bomb:think()
	Bomb.super.think(self)
	if not self:isActive() then return end
	local t = getTime() - self.boomTime
	self:setScale(vec2:new(math.cos(2*t*math.pi)*.2+.9))
	if t >= 0 then 	-- t minus ... like NASA time!
		self:explode()
		self:delete()
		return
	end
end

function makeSparks(pos)
	for i=1,10 do
		local c = math.random()
		Particle:new{
			vel = vec2:new(math.random()*2-1, math.random()*2-1);
			pos = pos;
			life = math.random() * .5 + .5;
			color = {1;c;c*math.random()*math.random();1};
			radius = .5 * math.random() + .25;
		}
	end
end

--[[
I need a custom uniform for this shader ... "starttime" ...
 and it only needs to be set once - when the object is created ...
 in the mean time, just use the circular light tex for our sprite, and just scale it for a sense of distortion
--]]
animLoad('bombwarp', 'lights/highres/circular_intense.bmp')	-- this is in base
entclass('BombWarp', Renderable)
BombWarp.animGroup = 'bombwarp'
BombWarp.shader = renderShader{ fragmentCode = advectionFragmentCode }
BombWarp.postFBO = true	-- remember kids, postFBO + last-screen-framebuffer effects = current-screen-framebuffer effects without the slowdown incurred from using this-screen-framebuffer effects
BombWarp.duration = .3
BombWarp.zBias = -10
function BombWarp:ctor(pos, maxRadius)
	BombWarp.super.ctor(self)
	self:setPos(pos)
	self.startTime = getTime()
	self.maxRadius = 10
end
function BombWarp:think()
	BombWarp.super.think(self)
	local t = (getTime() - self.startTime) / self.duration
	if t > 1 then t = 1 end
	self:setScale(vec2:new(t*self.maxRadius))
	if t == 1 then
		self:delete()
		return
	end
end

-- TODO for net optimization:
--  send across the network only the following (no particles)
--	x min, max
--	y min, max
--	pos is useful but it'll already be provided
--  particle seed or something
--  explosion start time
function Bomb:explode()
	-- start at our current position
	-- cycle in all directions
	local sides = { vec2:new(0,-1), vec2:new(0,1), vec2:new(-1,0), vec2:new(1,0) }
	BombWarp:new(self:getPos(), self.bombLength)
	makeSparks(self:getPos())
	for i,dir in ipairs(sides) do
		local pos = vec2:new(self:getPos())
		for j=0,self.bombLength do
			makeSparks(pos)
			-- if you want, we can trace as far as we can go, wrap a bbox around it, then just do one test to all ents
			-- if there's a guy within .5 linf then KILL THEM
			local hit = false	--bailout on hitting an entity
			for i,e in pairs(bombables) do
				if e ~= self then
					local snappos = vec2:new(e:getPos()):snap()
					-- if e snaps to the same tile that the bomb would then kill it
					if (pos - snappos):lenLInf() < .4 then
						if e.hitByBomb then
							e:hitByBomb()
							if e.blocksExplosions then
								hit = true
							end
						end
					end
				end
			end
			if hit then break end
			-- move
			pos:addTo(dir)
			-- if the map type is solid then stop
			if mapGetType(unpack(pos)) == MAPTYPE.Solid then break end
		end
	end
end

function Bomb:hitByBomb()
	-- give us a .1 second delay to die
	self.boomTime = getTime() + self.chainTime
end

