require 'bomblo.scripts.movable'		-- for shot esp
require 'bomblo.scripts.puff'
require 'bomblo.scripts.money'



-- all simulated in script.  robot style
entclass('MedusaShot', Movable)
function MedusaShot:ctor(owner)
	MedusaShot.super.ctor(self)
	self.owner = owner
	self:setPos(owner:getPos())		-- map dest as well
end

-- medusas shoot through trees
function MedusaShot:cannotPassThru(maptype)
	return maptype ~= MAPTYPE.Empty
		and maptype ~= MAPTYPE.Tree
		and maptype ~= MAPTYPE.Water
end

-- tell it to die when it hits the world border or maptype solid
-- tell it to die when it hits framers
-- tell it to die when it hits moneys
-- tell it to KILL if it hits a player
-- AND DONT STOP FOR DIRECTIONS (dont use test_object)
function MedusaShot:hitObject(what, pushDest, side)
	if what == self.owner then return self.HIT_RESPONSE.MOVE_THRU end
	if what:isa(Player) then
		what:die()
		return self.HIT_RESPONSE.STOP
	end
	if what.isBlocking or what:isa(Money) then return self.HIT_RESPONSE.STOP end			-- any blockable thing .. make a flag out of this?
	return self.HIT_RESPONSE.MOVE_THRU
end


animLoad('medusa', {
	sprites = {
		{file='medusa.png'};
		{file='medusa_pissed.png'};
	};
	seqs = {
		idle = {{joint = 1; time = 0; sprite = 1}};
		pissed = {{joint = 1; time = 0; sprite = 2}};
	}
})

entclass('Medusa', Base)	-- are medusas eggable?
Medusa.animGroup = 'medusa'
Medusa.bbox = box2:new(.4)
Spawnable(Medusa)
Medusa.FIRE_DIST = 0.25
Medusa.MAD_DIST = 0.75
function Medusa:think()
	Medusa.super.think(self)
	if not self:isActive() then return end

	-- if any players
	-- ... should I keep track of instances relative to their classes?
	-- unlink upon delete() ?
	self:setAnimSeq('idle')
	for i,e in pairs(_ents) do
		if e:isa(Player) and not e.dead then
	
			local diff = vec2:new(e:getPos()) - self:getPos()		-- vector from medusa to player
			local absDiff = diff:copy():abs()			-- absolute
			local dist = absDiff:minElem()		-- minimum element

			if dist < self.MAD_DIST then
				self:setAnimSeq('pissed')
			end

			if dist < self.FIRE_DIST then
				-- okay ...
				-- first we snap to 0.5 the medusa position
				-- then we calculate a delta vector to the player
				-- then we snap it to 90 degrees
				
				-- dont forget
				-- you either are standing
				-- or you are moving
				-- but never both
				
				local dir
				if diff[1] < diff[2] then		-- left or down
					if diff[1] < -diff[2] then dir = SIDE.LEFT else dir = SIDE.DOWN end
				else	-- up or right
					if diff[1] < -diff[2] then dir = SIDE.UP else dir = SIDE.RIGHT end
				end
				
				-- travel along it (just past the medusa) checking the left & right for blocking objects
				-- here's our OOP resource-inefficient but development-efficient solution:
				-- create a med.usa shot object
				-- move it thru the world
				-- AND DO ALL THIS manually
				--
				local shot = MedusaShot:new(self)	-- snap to half tile
				local response
				repeat
					response = shot:doMove(dir)
					shot:setPos(shot.destPos)
				until response == shot.MOVE_RESPONSE.WAS_BLOCKED
				shot:delete()
			end
		end
	end
end

function Medusa:onChestOpen()
	Puff(self:getPos())
	self:delete()
end
