-- notice to all require'rs of misc.lua:
-- you might be looking for base's misc.lua
-- in that case you should explicitly include that particular file
-- or, maybe, to make everyone happy, maybe i'll just include it here once and for all

require 'base.scripts.misc'

-- door.  this is specific to the 'data' mod
	
animLoad('door', {
	sprites = {
		{file='door/door_close.tga', bbox={-2,-2,2,2}};
		{file='door/door_pipe.tga', bbox={-1.5,-2,1.5,2}};
	};
	seqs = {
		idle = {{joint=1; time=0; sprite=1}};
		die = {{joint=1; time=0; sprite=2}};
	};
})

entclass('MetroidDoor', TakesDamage)
union(MetroidDoor,{

	animGroup = 'door';
	zBias = -7;	
	bbox = box2:new(-1,-2,1,2);

	closeBehindTime = 0.1,
	stayOpenTime = 5,
	STATE_CLOSED = {},
	STATE_OPEN = {},
	allowEnemyDeath = false,		-- override that which is in TakesDamage
	maxHealth = 2,
})
MetroidDoor.state = MetroidDoor.STATE_CLOSED
union(MetroidDoor,{
	doorCloseTime = 0;

	ctor = function(self,so)
		MetroidDoor.super.ctor(self, so)
		self:setMassInv(0)
	end,

	touch = function(self,other)
		MetroidDoor.super.touch(self,other)
		if not other or not other:isa(Player) then return end
		
		if self:getTouchOnly() then
			self.doorCloseTime = getTime() + self.closeBehindTime
		end
	end,

	death = function(self, other, amount)
		self:setAnimSeq('die')
		self.state = self.STATE_OPEN
		self.doorCloseTime = getTime() + self.stayOpenTime
	end,

	think = function(self)
		MetroidDoor.super.think(self)
		if not self:isActive() then return end	-- needed for any whose parent think() might kill themselves (Collidable)
		if self.state == self.STATE_CLOSED then
			self:solid()
		elseif self.state == self.STATE_OPEN then
			self:unsolid()
		end
	end,

	solid = function(self)
		self:setTouchOnly(false)
	end,

	unsolid = function(self)
		self:setTouchOnly(true)
		self.health = self.maxHealth
		
		if self.doorCloseTime < getTime() then
			self:setAnimSeq('idle')
			self.state = self.STATE_CLOSED
		end
	end,
})
Spawnable(MetroidDoor)


	-- super mario world springboard

animLoad('springboard', {
	sprites = {
		{file='springboard1.tga'};
		{file='springboard2.tga'};
		{file='springboard3.tga'};
	};
	seqs = {
		idle = {{joint=1; time=0; sprite=1}};
		spring = {
			{joint=1; time=0; sprite=2};
			{joint=1; time=.1; sprite=1};
			{joint=1; time=.3; sprite=2};
			{joint=1; time=.4; sprite=2};
		};
	};
})


entclass('Springboard', Collidable)

Springboard.animGroup = 'springboard'
Springboard.force = 10
Springboard.recoilTime = 0.5

function Springboard:ctor(so)
	Springboard.super.ctor(self,so)
	self.lastSpringTime = 0
	Springboard:readSpawnParams(self,so)
end

function Springboard:touch(other)
	Springboard.super.touch(self,other)
	if not other or other:getNoPhysics() then return end
	
	local gametime = getTime()
	if gametime - self.lastSpringTime < self.recoilTime then return end
	self.lastSpringTime = gametime
	
	other:setVelX(other:getVelX() + math.sin(self:getAngle() * math.pi / 180) * self.force)
	other:setVelY(other:getVelY() - math.cos(self:getAngle() * math.pi / 180) * self.force)
	
	self:runAnimSeq('spring')
end

SetupSpawnParams(Springboard, {
	force = 'force at launch';
	recoilTime = 'time taken to recoil';
})

Spawnable(Springboard)


entclass('TestObj', Collidable)
TestObj.animGroup = 'marker'
Spawnable(TestObj)
TestObj.friction = 1	--.5
--TestObj.restitution = 1
function TestObj:ctor(args)
	TestObj.super.ctor(self, args)
	print('ctor',self:getVelX(), self:getVelY())
end
function TestObj:touch(other)
	print('touch',self:getVelX(), self:getVelY())
end
function TestObj:think()
	TestObj.super.think(self)
	if math.abs(self:getVelX()) < .01 and math.abs(self:getVelY()) < .01 then
		self:delete()
	end
end
