require 'base.scripts.clientinput'
require 'base.scripts.hud'			-- drawRect
require 'base.scripts.clientview'
require 'ddiad.scripts.tile'
require 'ddiad.scripts.tile3x3'
require 'ddiad.scripts.board'
require 'ddiad.scripts.vars'

function newHandTile()
	-- enum of all possible tiles for random creation
	tileClasses = {Tile, Tile3x3}
	local tileClass = tileClasses[math.random(#tileClasses)]
	return tileClass:new()	-- no params means random content I guess
end

entclass('Player', ClientView)

Player:createNetTable{
	startTime = 'float';
	hasWon = 'bool';
	numMoves = 'int';
}

Player.ALWAYS_SHOW_TIMER = true
Player.animGroup = 'null'
Player.numMoves = 0
Player.startTime = 0
Player.hasWon = false
Player.numMoves = 0
	
function Player:ctor(so)
	Player.super.ctor(self,so)
	self.input = ClientInput:new()
	self:setPos{18,10}
	
	-- now prompt them through the menus and stuff
	-- if I'm going to be calling Menu:new then I need a clientview
	-- ... so I can create these clientside and do their updates and what not
	-- THAT'S RIGHT you CANNOT do send messages until after we've connected
	-- this takes a trick -- updating a flag on the client side -- 
	--	(the only place that we use clientside writes
	--		consider incorperating this right into the ent system and getting rid of client writes)
	fork(function()
		while not self.clientAck do sleep() end
		self:serverSendMessage('clientCreateMainMenu')
	end)
end

function Player:serverStartGame(boardsize)
	boardsize = tonumber(assert(boardsize))
	print('boardsize',boardsize)
	self:startGame(boardsize)
end

-- 'net' refers to the remote entity, 'self' is the ClientSide object which is called by the class method 'clientUpdate' of the class type that the player server entity is
function ClientSide:updateTimer()
	if self.timer and not net.hasWon then
		self.timer.text =
			'seconds: '..math.floor(getTime() - net.startTime)..'\n'..
			'moves: '..net.numMoves
	end
end

function ClientSide:startGame(boardsize)
	-- tell the server to start
	entClientSendMessage(self.entID, 'serverStartGame', boardsize)
	
	-- start ourselves

	if Player.ALWAYS_SHOW_TIMER then
		-- only starrt the timer once everything is set up
		self.timer = Menu:new{
			client=self;
			pos={1,1};
			size={30,4};
			fontSize={2,2};
		}
		self.updateCallbacks:insertUnique(ClientSide.updateTimer)
	end
end

function ClientSide:clientCreateMainMenu()
	local ms = table.new()

	local function clearStartMenu()
		for _,m in ipairs(ms) do
			m:delete()
		end
	end

	for i=1,3 do
		local boardsize = i+2
		local m = Menu:new{
			client = self;
			pos = {20,30+(i-1)*3};
			size = {20,2};
			text = boardsize..'x'..boardsize..' Board';
			fontSize = {2,2};
			boardsize = boardsize;
			mouseEvent = function(self, input)
				if input.key.attack.isReleased then
					self.client:startGame(self.boardsize)
					clearStartMenu()
				end
			end;
		}
		ms:insert(m)
	end
end

-- 
function Player:startGame(boardSize)
	
	-- while you're here, how about setting up the player's board too?
	self.board = Board:new{owner=self, size=boardSize}
	
	-- then the player's hand
	self.hand = table.new()
	for j=1,handHeight do
		for i=1,handSize do
			local p = Place:new()
			p:setPos((vec2:new(i,5-j+1) - vec2:new(handSize/2, 0)) * tileSize + self:getPos())
			self.hand:insert(p)
		end
	end
	
	-- and fill the player's hand with random cards!
	for i,p in pairs(self.hand) do
		newHandTile():setOn(p)
	end

	self.startTime = getTime()
end
	
function Player:think()
	Player.super.think(self)
	if not self:isActive() then return end
	
	local mouseWorldPos = vec2:new(self.input.mouse.pos):map(viewGetBox())
	
	--print(mouseWorldPos, self.input.key.attack.isDown, self.input.key.attack.isPressed, self.input.key.attack.isReleased)
	
	-- if nothing is held then
	if not self.held then
		if not self.hasWon then
			-- see if we clicked
			if self.input.key.attack.isPressed then
				-- see if its on a tile
				for e in allents() do
					if e:isa(Tile)
					and (not e.place or not e.place.isABoardPlace)
					and boxToWorld(e):contains(mouseWorldPos) then
						-- then pick it up
						self:acquireHeld(e)
						-- and only it
						break
					end
				end
			end
		end
	else -- if something is held then
		-- upate its position every frame
		self.held:setPos(mouseWorldPos)

		-- see if the user let up on the mouse
		if self.input.key.attack.isUp then
			-- see if there is a place under our tile
			for e in allents() do
				if e:isa(Place)
				and e.isABoardPlace
				and boxToWorld(e):contains(mouseWorldPos)
				then
					-- first see if playing it fits all constraints
					-- if it cant should we make a beep or somethin?
					if self.held:canPlay(e) then
					
						self.numMoves = self.numMoves + 1
					
						-- oh while we're here we can make some particles or something
						for i=1,30 do
							Particle:new{
								pos = vec2:new(self.held:getPos());
								vel = vec2.expd(math.random()*360):mulTo(10);
								life = .4;
								color = (vec4:new(self.held:getColor()) + .1):clamp(0,1);
								radius = 2;
							}
						end
					
						-- for now just replace whatever was there with a solid color
						if e.tile then e.tile:delete() end
						Tile:new(self.held.colorObj):setOn(e)

						-- remove it from its place in our hand
						local oldplace = self.held.place
						self.held:setOn(nil)	

						-- and get rid of what was being held
						self.held:delete()
						self.held = nil
						
						-- see if the board is one solid color
						self:checkWin()
						
						-- put a new one in our hand where it came from
						newHandTile():setOn(oldplace)
						
						break
					end
				end
			end
			
			-- if we are still holding onto it
			if self.held then
				-- if so move the tile back where it belongs
				self.held:moveToPlace()
				-- and let go of it
				self:releaseHeld()
			end
		end
	end
end
	
function Player:acquireHeld(e)
	assert(not self.held)
	self.held = e
	self.held.depth = -7
end
	
function Player:releaseHeld()
	if not self.held then return end
	-- reset all params
	self.held.depth = -6
	self.held = nil
end
	
function Player:checkWin()
	if not self.hasWon then
		local won = self.board:hasWon()
		
		if won then
			self.hasWon = true	-- flag so we don't make too many windows
			fork(function()
				local seconds = getTime() - self.startTime
				seconds = math.floor(seconds * 100 + .5) / 100 
				local minutes = math.floor(seconds / 60)
				seconds = seconds - minutes * 60
				local msg = 'YOU WON!!!\n'
				msg = msg .. 'time taken: '
				if minutes > 0 then
					msg = msg .. minutes .. ' minutes, '
				end
				msg = msg .. seconds .. ' seconds'
				self:dialog(msg)
			end)
		end
	end
end

-- no spawn type association required - its all done through the gameapi
-- in fact, no spawn stuff is done at all

