require 'base.scripts.ents'
require 'ddiad.scripts.vars'

buttonMaskTex = spriteLoad('buttonmask.png')
buttonOverlayTex = spriteLoad('buttonoverlay.png')

entclass('Tile', Renderable)
Tile.bbox = box2:new(.4);
function Tile:ctor(colorObj)
	Tile.super.ctor(self)
	-- for random creation
	if not colorObj then colorObj = colors[math.random(colorsUsed)] end
	--for renderable
	self:setScale(tileSize)
	self:setTileColor(colorObj)
end
function Tile:setTileColor(co)	-- co may be a color object or may be a tile
	if co.isa and co:isa(Tile) then
		co = co.colorObj
	end --assume its a color object
	
	self:setColor(co.color)
	self.colorObj = co
end
Tile.boardSize = 0
function Tile:setOn(place)
	-- out with the old
	if self.place then		-- if we already have a tile
		assert(self.place.tile == self)	-- then it better have us too
		self.place.tile = nil
	end
	-- in with the new
	if place then
		if place.tile then	-- delete when we replace?
			place.tile:delete()
		end
		place.tile = self
		if place.boardPos then
			self.boardPos = place.boardPos:copy()
		else
			self.boardPos = vec2:new{0,0}
		end
		if place.board then
			self.boardSize = place.board.size
		end
	else
		self.boardPos = vec2:new{0,0}
	end
	self.place = place
	-- and the sprite
	self:moveToPlace()
end
function Tile:moveToPlace()
	if self.place then
		self:setPos(self.place:getPos())
	end
end
-- for the solid tile version canplay always works
-- for other subclasses override it and override the renders as well
function Tile.canPlay(place) return true end

Tile.depth = -6

function Tile.clientUpdate(classObj)
	local b = boxToWorld(net, classObj.bbox)
	classObj:renderTile(b)
end

function Tile.renderTile(classObj, b)
	-- minus is closer to the viewer
	-- plus is deeper / further
	classObj.drawButton{
		box=b;
		color=net.color;
		depth=net.depth;
	}
	
	local faded = vec4:new(net.color[1], net.color[2], net.color[3], .5)

	if net.boardPos[1] == 1 then
		classObj.drawButton{
			box=box2:new(b) + vec2:new{net.boardSize * net.scale[1], 0};
			color=faded;
			depth=net.depth;
		}

		if net.boardPos[2] == 1 then
			classObj.drawButton{
				box=box2:new(b) + vec2:new{net.boardSize * net.scale[1], net.boardSize * net.scale[2]};
				color=faded;
				depth=net.depth;
			}
		end

		if net.boardPos[2] == net.boardSize then
			classObj.drawButton{
				box=box2:new(b) + vec2:new{net.boardSize * net.scale[1], -net.boardSize * net.scale[2]};
				color=faded;
				depth=net.depth;
			}
		end
	end

	if net.boardPos[1] == net.boardSize then
		classObj.drawButton{
			box=box2:new(b) + vec2:new{-net.boardSize * net.scale[1], 0};
			color=faded;
			depth=net.depth;
		}

		if net.boardPos[2] == 1 then
			classObj.drawButton{
				box=box2:new(b) + vec2:new{-net.boardSize * net.scale[1], net.boardSize * net.scale[2]};
				color=faded;
				depth=net.depth;
			}
		end

		if net.boardPos[2] == net.boardSize then
			classObj.drawButton{
				box=box2:new(b) + vec2:new{-net.boardSize * net.scale[1], -net.boardSize * net.scale[2]};
				color=faded;
				depth=net.depth;
			}
		end
	end

	if net.boardPos[2] == 1 then
		classObj.drawButton{
			box=box2:new(b) + vec2:new{0,net.boardSize * net.scale[2]};
			color=faded;
			depth=net.depth;
		}
	end

	if net.boardPos[2] == net.boardSize then
		classObj.drawButton{
			box=box2:new(b) + vec2:new{0,-net.boardSize * net.scale[2]};
			color=faded;
			depth=net.depth;
		}
	end
end
	
--static - i.e. no self
-- args: box, color, depth
function Tile.drawButton(args)
	local del = args.box:size()*.2
	drawRect{
		--color=args.color,
		color={0,0,0,1};
		box=box2:new{min=args.box.min-del, max=args.box.max+del};
		depth=args.depth+.1,
		tex=buttonMaskTex,
		blend=BLEND.DECAL
	}
	drawRect{
		color=args.color,
		--color=vec4:new(1,1,1, args.color[4]),
		box=args.box,
		depth=args.depth,
		tex=buttonOverlayTex,
		blend=BLEND.DECAL
	}
end

Tile:createNetTable{
	depth = 'float';	-- this is really only ever used as a flag for whether it's being picked up or not
	boardPos = 'vec2f';
	boardSize = 'float';
}
