/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <iostream>
#include "system/exception.h"
#include "system/singletons.h"
#include "base/base.h"
#include "menu/lib.h"

#include "inspiration/main.h"

#include "menu.h"
#include "impl.h"
#include "func.h"

//i was too lazy to switch it all so soon from win32
enum {
	ID_EDIT_SNAPTOGRID,
	ID_EDIT_HIDE_ENTS,
	ID_EDIT_RESIZE,
	ID_EDIT_MUNDO,
	ID_EDIT_MREDO,
	ID_FILE_NEW_WORLD,
	ID_FILE_OPEN_WORLD,
	ID_FILE_SAVE_WORLD,
	ID_FILE_SAVE_WORLD_AS,
	ID_FILE_EXIT_WORLD,
	ID_GAME_START,
	ID_GAME_PAUSE,
	ID_GAME_RESET_LEVEL,
	ID_GAME_RESET_WHOLE,
	ID_METHOD_EDIT_TAGS,
	ID_METHOD_SURFACEPAINT,
	ID_METHOD_SURFACEFILL,
	ID_METHOD_SURFACEMOVE,
	ID_METHOD_ENTITYCREATE,
	ID_METHOD_ENTITYSELECT,
	ID_WINDOW_WORLD,
	ID_WINDOW_TAG,
	ID_WINDOW_ENTITY,
	ID_WINDOW_TILE,
	ID_WINDOW_TEXTURE,
	ID_WINDOW_SCRIPT,
};


using namespace pwn;
using namespace std;

// ************************* EditorBaseMenu ************************* 

class EditorSubMenuEntry;
class EditorSubMenu : public SubMenu {
	typedef SubMenu super;
public:
	EditorSubMenu() {}
	
	virtual bool display();
	virtual void postDisplay();
	virtual eventResponse_t preMouseEvent(MouseEvent e, const vec2f &v);
};

class EditorSubMenuEntry : public SubMenuEntry {
	typedef SubMenuEntry super;
protected:
	EditorImpl *ed;
public:
	EditorSubMenuEntry(EditorImpl *ed_, const char *name, int id) : ed(ed_), super(name,id) {}
	
	virtual bool display() {
		if (!editor.isActive()) return false;
		if (!super::display()) return false;
		return true;
	}
	virtual void postDisplay() {
		if (!editor.isActive()) return;
		super::postDisplay();
	}
	virtual eventResponse_t preMouseEvent(MouseEvent e, const vec2f &v) {
		if (!editor.isActive()) return RESPONSE_STOP;
		return super::preMouseEvent(e, v);
	}
	
	//onclick anyone?
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v) {
		if (!e.lpress) return RESPONSE_CONTINUE;

		switch (ident) {
		case ID_FILE_NEW_WORLD:		ed->func.newWorld();		break;
		case ID_FILE_OPEN_WORLD:	ed->func.loadWorld();		break;
		case ID_FILE_SAVE_WORLD:	ed->func.saveWorld();		break;
		case ID_FILE_SAVE_WORLD_AS:	ed->func.saveWorldAs();		break;
		
		case ID_FILE_EXIT_WORLD:	sys.requestExit();						break;

		//case ID_EDIT_MUNDO:			ed->func.undo();						break;
		//case ID_EDIT_MREDO:			ed->func.redo();						break;
		case ID_EDIT_RESIZE:		ed->func.resize();					break;
		case ID_EDIT_SNAPTOGRID:	editor->snapEnabled ^= 1;	break;
		case ID_EDIT_HIDE_ENTS:	editor->hideEnts ^= 1;	break;

		case ID_GAME_START:			gameState = GS_RUNNING;				break;
		case ID_GAME_PAUSE:			gameState = GS_PAUSED;					break;
		case ID_GAME_RESET_LEVEL:	mainSys.sessionReset();				break;
		case ID_GAME_RESET_WHOLE:	mainSys.gameReset();				break;

		case ID_METHOD_EDIT_TAGS:	editor->setMethod(EDITOR_EDIT_TAGS);		break;
		case ID_METHOD_SURFACEPAINT:editor->setMethod(EDITOR_SURFACE_PAINT);	break;
		case ID_METHOD_SURFACEFILL:	editor->setMethod(EDITOR_SURFACE_FILL);	break;
		case ID_METHOD_SURFACEMOVE:	editor->setMethod(EDITOR_SURFACE_MOVE);	break;
		case ID_METHOD_ENTITYCREATE:editor->setMethod(EDITOR_ENT_CREATE);	break;
		case ID_METHOD_ENTITYSELECT:editor->setMethod(EDITOR_ENT_SELECT);	break;

		case ID_WINDOW_WORLD:		ed->func.showWindow(EW_MAIN, 2);		break;
		case ID_WINDOW_TAG:			ed->func.showWindow(EW_TAG, 2);		break;
		case ID_WINDOW_ENTITY:		ed->func.showWindow(EW_ENT, 2);		break;
		case ID_WINDOW_TILE:		ed->func.showWindow(EW_TILE, 2);		break;
		case ID_WINDOW_TEXTURE:		ed->func.showWindow(EW_TEX, 2);		break;
		case ID_WINDOW_SCRIPT:		ed->func.showWindow(EW_SCRIPT, 2);	break;
		}
		
		//hide here
		//...somehow
		
		return RESPONSE_CONTINUE;
	}
};

bool EditorSubMenu::display() {
	if (!editor.isActive()) return false;
	if (!super::display()) return false;
	return true;
}
void EditorSubMenu::postDisplay() {
	if (!editor.isActive()) return;
	super::postDisplay();
}
eventResponse_t EditorSubMenu::preMouseEvent(MouseEvent e, const vec2f &v) {
	if (!editor.isActive()) return RESPONSE_STOP;
	return super::preMouseEvent(e, v);
}


class EditorBaseMenu : public BaseMenu {
	typedef BaseMenu super;
protected:
	EditorImpl *ed;
	
public:

	void addMenu(const char *name, EditorSubMenu *m) {
		addChild(new BaseMenuEntry(name, m, editor->rootmenu));
	}

	EditorBaseMenu(EditorImpl *ed_) : ed(ed_) {
		{
			EditorSubMenu *file = new EditorSubMenu();
			file->addChild(new EditorSubMenuEntry(ed, "New", ID_FILE_NEW_WORLD));
			file->addChild(new EditorSubMenuEntry(ed, "Open...", ID_FILE_OPEN_WORLD));
			file->addChild(new EditorSubMenuEntry(ed, "Save", ID_FILE_SAVE_WORLD));
			file->addChild(new EditorSubMenuEntry(ed, "Save As...", ID_FILE_SAVE_WORLD_AS));
			file->addChild(new EditorSubMenuEntry(ed, "", 0));			//separator
			file->addChild(new EditorSubMenuEntry(ed, "Exit", ID_FILE_EXIT_WORLD));
			addMenu("File", file);
		}
		{
			EditorSubMenu *edit = new EditorSubMenu();
			edit->addChild(new EditorSubMenuEntry(ed, "Undo", ID_EDIT_MUNDO));
			edit->addChild(new EditorSubMenuEntry(ed, "Redo", ID_EDIT_MREDO));
			edit->addChild(new EditorSubMenuEntry(ed, "", 0));		//separator
			edit->addChild(new EditorSubMenuEntry(ed, "Resize...", ID_EDIT_RESIZE));
			edit->addChild(new EditorSubMenuEntry(ed, "", 0));		//separator
			edit->addChild(new EditorSubMenuEntry(ed, "Snap to Grid", ID_EDIT_SNAPTOGRID));
			edit->addChild(new EditorSubMenuEntry(ed, "", 0));		//separator
			edit->addChild(new EditorSubMenuEntry(ed, "Hide Ents", ID_EDIT_HIDE_ENTS));
			addMenu("Edit", edit);
		}
		{
			EditorSubMenu *gameMenu = new EditorSubMenu();
			gameMenu->addChild(new EditorSubMenuEntry(ed, "Start", ID_GAME_START));
			gameMenu->addChild(new EditorSubMenuEntry(ed, "Pause", ID_GAME_PAUSE));
			gameMenu->addChild(new EditorSubMenuEntry(ed, "Reset Level", ID_GAME_RESET_LEVEL));
			gameMenu->addChild(new EditorSubMenuEntry(ed, "Reset Game", ID_GAME_RESET_WHOLE));
			addMenu("Game", gameMenu);
		}
		{
			EditorSubMenu *method = new EditorSubMenu();
			method->addChild(new EditorSubMenuEntry(ed, "Edit Tags", ID_METHOD_EDIT_TAGS));
			method->addChild(new EditorSubMenuEntry(ed, "", 0));
			method->addChild(new EditorSubMenuEntry(ed, "Surface Paint", ID_METHOD_SURFACEPAINT));
			method->addChild(new EditorSubMenuEntry(ed, "Surface Fill", ID_METHOD_SURFACEFILL));
			method->addChild(new EditorSubMenuEntry(ed, "Surface Move", ID_METHOD_SURFACEMOVE));
			method->addChild(new EditorSubMenuEntry(ed, "", 0));
			method->addChild(new EditorSubMenuEntry(ed, "Entity Create", ID_METHOD_ENTITYCREATE));
			method->addChild(new EditorSubMenuEntry(ed, "Entity Select", ID_METHOD_ENTITYSELECT));
			addMenu("Method", method);
		}
		{
			EditorSubMenu *window = new EditorSubMenu();
			window->addChild(new EditorSubMenuEntry(ed, "World", ID_WINDOW_WORLD));
			window->addChild(new EditorSubMenuEntry(ed, "Tag", ID_WINDOW_TAG));
			window->addChild(new EditorSubMenuEntry(ed, "Tile", ID_WINDOW_TILE));
			window->addChild(new EditorSubMenuEntry(ed, "Texture", ID_WINDOW_TEXTURE));
			window->addChild(new EditorSubMenuEntry(ed, "Entity", ID_WINDOW_ENTITY));
			//new EditorSubMenuEntry("Script", ID_WINDOW_SCRIPT)	//no more script windows... thought a listing would be nice
			addMenu("Window", window);
		}
	}

	virtual bool display() {
		if (!editor.isActive()) return false;
		return super::display();
	}
	virtual void postDisplay() {
		if (!editor.isActive()) return;
		super::postDisplay();
	}

	virtual eventResponse_t preMouseEvent(MouseEvent e, const vec2f &v) {
		if (!editor.isActive()) return RESPONSE_STOP;
		return super::preMouseEvent(e, v);
	}
};

// ************************* end menu classes ************************* 

void EditorMenu::init() {
	EditorBaseMenu *m = new EditorBaseMenu(ed);
	if (!m) throw Exception() << "EditorMenu::init failed to create EditorBaseMenu";
//virtually no slowdown from the added menus
	editor->rootmenu->addChild(m);
}

EditorMenu::EditorMenu(EditorImpl *ed_) : ed(ed_) {}
