/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <stdio.h>

#include "system/exception.h"
#include "base/base.h"

#include "inspiration/main.h"
#include "inspiration/game/main.h"

#include "wndent.h"
#include "impl.h"

using namespace std;
using namespace pwn;

//don't use the text listener cuz we need access to onblur/onlisten
//i suppose i could make some focus listeners ...
//and i'im using this 'lasttext' field too ...
class EntTextField : public TextField {
	typedef TextField super;
public:
	EntTextField(const vec2f &_pos, float _width) : super(_pos, _width) {}
	
	string lasttext;
	
	virtual void onBlur();
	virtual void onFocus();
	virtual void onTextChange();	
};

class MenuEnt : public MenuEditorWindow, public SelectListener, public TextListener {
	typedef MenuEditorWindow super;
public:

	gc::Pointer<Text> methodCreate;
	gc::Pointer<Text> methodSelect;

	gc::Pointer<Text> entityDelete;
	gc::Pointer<Text> entityTitle;	//"entity ###"

	gc::Pointer<Select> selectSpawnType;

	gc::Pointer<TextScrollContainer> entityDesc;	//todo - dialog box style
	
	gc::Pointer<Text> add;
	gc::Pointer<Text> del;
	gc::Pointer<Text> spawn;
	
	gc::Pointer<EntTextField> name;

	gc::Pointer<ScrollContainer> varPanel;	//list containing key/val's
	
	MenuEnt();
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
	virtual bool display();
	
	virtual void callback_userSetSelected(Select *, int newSel, int oldSel);
	virtual void callback_onTextChange(TextField *m);
};

//********************
//ENTITY EDITOR WINDOW
//********************

//editorWndEnt.refreshTypeTable is only called
//upon change level
//and before the type code is computed
//so it only gets the last levels type code
//and it gets it repeated as many times as the user ran the level
//(because types dont search for uniqueness when they're added)
//(nor are they cleared when levels are reset)
/*
THIS METHOD IS CALLED BY editorWndEnt.refresh()
WHICH IS ONLY CALLED BY Editor_RefreshAll()
Editor_RefreshAll is called by ieditor.refreshAll
ieditor.refreshAll is called just before progSys.gameInit() at first
 ... which does nothing
we really need it always called after the level load
because thats what will refresh the typenames ...

what about users starting new levels?
haven't thought that far ahead quite yet ...

dont refresh the editor just after loading the map
(that gives it textures at most)
call it just after running the basecode
(this will give it types too)

*/


class EditorWndEntImpl {
public:
	gc::GlobalPointer<MenuEnt> entMenu;

	EditorWndEntImpl() : entMenu(gc::sys->createPtr<MenuEnt>()) {}

	void init();
	void refresh();
	void refreshSelected();
	void refreshTypeCurrent();
	void refreshTypeTable();
	void refreshTypeText();
	void refreshName();
	void addListVar(const char *key, const char *value);
	void refreshListVars();

};

// ************************* MenuEnt ************************* 

MenuEnt::MenuEnt() 
:	super(vec2f(1,1),vec2f(27,25)),
	methodCreate(createPtr<Text>()),
	methodSelect(createPtr<Text>()),
	entityDelete(createPtr<Text>()),
	entityTitle(createPtr<Text>()),
	selectSpawnType(createPtr<Select>()),
	entityDesc(createPtr<TextScrollContainer>()),
	add(createPtr<Text>()),
	del(createPtr<Text>()),
	spawn(createPtr<Text>()),
	name(createPtr<EntTextField>()),
	varPanel(createPtr<ScrollContainer>())
{
	visible = false;
	methodCreate = new Text(vec2f(1,1), "Create"); addChild(methodCreate);	//on-click create a new entity 
	methodSelect = new Text(vec2f(5,1), "Select"); addChild(methodSelect);	//on-drag select entities
	entityDelete = new Text(vec2f(10,1), "Delete"); addChild(entityDelete);	//delete all currently-selected ents
	spawn = new Text(vec2f(15,1), "Spawn"); addChild(spawn);			//spawn the entity
	
	entityTitle = new Text(vec2f(1,3), "Entity X"); addChild(entityTitle);	//which ent is selected?
	addChild(new Text(vec2f(1,4), "Name:"));
	name = new EntTextField(vec2f(5,4), 10); addChild(name);			//whats its name?
	addChild(new Text(vec2f(1,5), "Class:"));
	selectSpawnType = new Select(vec2f(5,5), 17); addChild(selectSpawnType);		//whats its type?
	selectSpawnType->addSelectListener(this);
	selectSpawnType->addTextListener(this);
	selectSpawnType->allowEdit = true;
	
	entityDesc = new TextScrollContainer(vec2f(0,0), vec2f(27,30));
	addChild(new ScrollArea(vec2f(1,7), vec2f(25,8), entityDesc));
	
	add = new Text(vec2f(1,16), "Add"); addChild(add);		//add a variable
	del = new Text(vec2f(5,16), "Del"); addChild(del);	//remove the currently selected

	//TODO - scrollbars for the varPanel's scrollarea aren't accepting clicks...
	varPanel = new ScrollContainer(vec2f(0,0), vec2f(24, 4));
	varPanel->setBorderColor(0);
	addChild(new ScrollArea(vec2f(1,17), vec2f(25, 6), varPanel));
}

eventResponse_t MenuEnt::mouseEvent(MouseEvent e, const vec2f &v) {
	if (e.lpress) {
		if (selectedChild.ptr() == methodCreate) {
			editor->setMethod(EDITOR_ENT_CREATE);
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == methodSelect) {
			editor->setMethod(EDITOR_ENT_SELECT);
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == entityDelete) {
			editor->deleteEntity();
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == spawn) {
			for (int i = 0; i < spawnObjSys.count(); i++) {
				SpawnObj *s = spawnObjSys[i];
				if (!s) {
					con.print("got invalid ptr for index %d of %d\n", i, spawnObjSys.count());
					continue;
				}
				if (!s->sel) continue;

				entptr_t e = s->ent;
				if (e.i != -1) {
					game.deleteEntity(e);
				}

				if (!s->createEnt()) {
					con.print("IDC_BUTTON_SPAWN: spawnObjCreateEntity failed on object %d\n", i);
				}
			}
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == add) {
			editorWndEnt->addListVar("", "");
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == del) {
			if (editor->spawnObj) {
//				spawnObj->m.erase(
//					&spawnObj->m.begin()[spawnPair]);

				//instead of using the iterator (which doesnt match up with the map)
				//how about we use the map and try removing the specified key?

				int j = 0;
				for (list<gc::Pointer<Object> >::iterator mi = varPanel->getKids().begin(); mi != varPanel->getKids().end(); ++mi, ++j) {
					Object *m = mi->ptr();
					if (j == editor->spawnPair<<1) {
						map<string,string>::iterator iter = editor->spawnObj->m.find(m->getText());
						if (iter != editor->spawnObj->m.end()) {
							editor->spawnObj->m.erase(iter);
						}
						break;
					}
				}


				editorWndEnt.refreshSelected();
			}
			return RESPONSE_STOP;
		}
	}
	
	//do sliders here?
	
	super::mouseEvent(e, v);
	return RESPONSE_STOP;
}

bool MenuEnt::display() {
	//update colors
	methodCreate->setFontColor((editor->method == EDITOR_ENT_CREATE) ? 0xFF0000FF : 0xFFFFFFFF);
	methodSelect->setFontColor((editor->method == EDITOR_ENT_SELECT) ? 0xFF0000FF : 0xFFFFFFFF);
	//display rect
	return super::display();
}

//architecture question:
//should onTextChange be called on editoable select objects when their selection is changed without using text?
//because the result would be a text change, however the user would not have typed anything...
void MenuEnt::callback_onTextChange(TextField *m) {
	if (m != selectSpawnType) return;

	editor->spawnType = m->getText();

	if (editor->spawnObj) {
		editor->spawnObj->rw("classname") << editor->spawnType;
		editorWndEnt->refreshTypeText();
	}
}

void MenuEnt::callback_userSetSelected(Select *select, int newSel, int oldSel) {
	if (select != selectSpawnType) return;
	if (newSel == oldSel) return;

	editor->spawnType = "";
	if (newSel >= 0 && newSel < select->items.size()) {
		editor->spawnType = select->items[newSel]->getText();
	}
	//spawnType = spawnTypeSys.get(selectedIndex);

	if (editor->spawnObj) {
		editor->spawnObj->rw("classname") << editor->spawnType;
		editorWndEnt->refreshTypeText();
	}
}

// ************************* EntTextField ************************* 

//if the user clicks a member of the varPanel then set the appropriate spawnObj
void EntTextField::onFocus() {
	super::onFocus();
	if (parent.ptr() == editorWndEnt->entMenu->varPanel) {
	
	//backup the current text - so onTextEdit if we screw up we got something
		lasttext = text;
	
		//an entity field:
//		first get the child's index
		int si = parent->indexOfChild(this) >> 1;
		if (editor->spawnObj) {
//			ESpawnPair *pair = spawnObj->getPair(si);
//			if (pair) {	//if we got a valid pair
				editor->spawnPair = si;
//			}
		}
	}
}

void EntTextField::onBlur() {
	super::onBlur();
	if (parent.ptr() == editorWndEnt->entMenu->varPanel) {
		editor->spawnPair = -1;
	}
}

void EntTextField::onTextChange() {
	super::onTextChange();
	if (this == editorWndEnt->entMenu->name) {
		if (editor->spawnObj &&
			editor->nameIsValid(text.c_str()))
		{
			editor->spawnObj->rw("name") << text;
		}
	} else if (parent.ptr() == editorWndEnt->entMenu->varPanel) {
#if	1
		//well maps wont do so well when we try to insert two of the same keys
		//and i could see a user doing just that
		//so my solution? any time a thing is changed we go through and upload THEM ALL
		editor->spawnObj->m.clear();
		for (list< gc::Pointer<pwn::Object> >::iterator i = editorWndEnt->entMenu->varPanel->getKids().begin(); i != editorWndEnt->entMenu->varPanel->getKids().end(); ++i) {
			string key = (*i)->getText();
			++i;
			if (i == editorWndEnt->entMenu->varPanel->getKids().end()) break;
			Object *m = *i;
			//let the for loop do the next set and test ...
			editor->spawnObj->m[key] = m->getText();
		}
		
#else
		//an entity field:
//		first get the child's index
		int i = 0;
		Object *m = prev;
		for (; m; m=m->prev, i++);	//count how many
		int si = i >> 1;	//get the spawn index
		if (si >= 0 && si < spawnObj->m.size()) {
//		ESpawnPair *pair = spawnObj ? spawnObj->getPair(si) : NULL;
//		if (pair) {	//if we got a valid pair
			spawnPair = si;

	//		if its mod 0 then we've edited a keyname - find the matching key - if it doesnt exist then . . . hmm . . . fail?  what was the old text anyways?
			if (!(i & 1)) {
				//refresh the old text
				//assert(pair->key);	//this is failing . . .
				//strncpy(text, pair->key->name, sizeof(text));
				text = lasttext;
				return;
			}
	//		if its mod 1 then we've edited a value - upload it!
			else {
//				pair->setValueString(text);
				map<string,string>::iterator iter = spawnObj->m.begin();
				for (int qtw = si; qtw > 0; qtw--) iter++;
				iter->second = text;
con << " setting index " << si << " key " << iter->first << " to " << text << "\n";
			}
		} else {
			//con.print("entMenu key/value editor didnt find a key under index %d\n", si);
			if (spawnObj) {
				//didnt get an existing pair - try adding a new pair:
				spawnPair = si;
				if (!(i & 1)) {
con << " erasing " << text << "\n";
					spawnObj->m[text] = "";
				}
			}
		}
#endif
	}
}

// ************************* EditorWndEntImpl ************************* 

void EditorWndEntImpl::init() {
	(*editor->editWnd[EW_ENT]) = entMenu = new MenuEnt();
	if (!entMenu) throw Exception() << "EditorWndEnt::init failed";
	editor->rootmenu->addChild(entMenu);
}

//exec when the selected object changes
void EditorWndEntImpl::refreshSelected() {
	//the label
	char buffer[256];
	if (!editor->spawnObj) {
		strcpy(buffer, "No Entity Selected");
	} else {
		sprintf(buffer, "Entity %d", spawnObjSys.getIndex(editor->spawnObj));
	}
	entMenu->entityTitle->setText(buffer);
	entMenu->entityTitle->setSize(vec2f(entMenu->entityTitle->getText().length(), 1));

	//selected name
	refreshName();

	//and the vars
	refreshListVars();

	//type current:
	editor->spawnType = "";
	if (editor->spawnObj) editor->spawnObj->rw("classname") >> editor->spawnType;
	
	refreshTypeCurrent();
}

void EditorWndEntImpl::refreshTypeCurrent() {
	entMenu->selectSpawnType->setSelected(
		spawnTypeSys.getIndexByName(editor->spawnType.c_str())
	);
	refreshTypeText();
}

void EditorWndEntImpl::refreshTypeTable() {
	entMenu->selectSpawnType->resetContent();
	for (int i = 0; i < spawnTypeSys.count(); i++) {
		entMenu->selectSpawnType->addString(spawnTypeSys.getNameForIndex(i));
	}
}

void EditorWndEntImpl::refreshTypeText() {
	entMenu->entityDesc->setText(
		spawnTypeSys.getDescByName(
			editor->spawnType.c_str()
		)
	);
}

void EditorWndEntImpl::refreshName() {
	string s = "";
	if (editor->spawnObj) editor->spawnObj->rw("name") >> s;
	entMenu->name->setText(s);
}

void EditorWndEntImpl::addListVar(const char *key, const char *value) {

	//count the # of children
	int i = entMenu->varPanel->getKids().size() >> 1;

	EntTextField *t;

	t = new EntTextField(vec2f(0, i), 9);
	t->setText(key);
	//t->setBorderColor(0);
	entMenu->varPanel->addChild(t);

	t = new EntTextField(vec2f(11, i), 24);
	t->setText(value);
	//t->setBorderColor(0);
	entMenu->varPanel->addChild(t);
	
	float sy = i + 1;
	if (sy < entMenu->varPanel->getParent()->getSize().y - 1)
		sy = entMenu->varPanel->getParent()->getSize().y - 1;
	
	entMenu->varPanel->setSize(vec2f(37, sy));
}

//as we go, add pairs of key/value EntTextField to varPanel
void EditorWndEntImpl::refreshListVars() {
	char buffer[256];

	//remove all its children
	entMenu->varPanel->getKids().clear();
	
	//no ones selected - leave it empty
	if (!editor->spawnObj) return;

	for (map<string,string>::iterator iter = editor->spawnObj->m.begin();
		iter != editor->spawnObj->m.end(); iter++)
	{
		addListVar(iter->first.c_str(), iter->second.c_str());
	}
	//select ? wha?  current spawn pair? . . . like it matters ?
}


void EditorWndEntImpl::refresh() {

	//refresh the 'type' list
	refreshTypeTable();

//	//select the current ent type
//	//now within RefreshSelected
//	refreshTypeCurrent();

	//refresh the current ent properties
	refreshSelected();
}

// ************************* EditorWndEnt ************************* 

EditorWndEnt::EditorWndEnt() : impl(new EditorWndEntImpl()) {}
EditorWndEnt::~EditorWndEnt() { delete impl; }
void EditorWndEnt::init() { impl->init(); }
void EditorWndEnt::refresh() { impl->refresh(); }
void EditorWndEnt::refreshSelected() { impl->refreshSelected(); }
void EditorWndEnt::refreshTypeCurrent() { impl->refreshTypeCurrent(); }
EditorWndEnt editorWndEnt;
