/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <stdio.h>
#include <stdlib.h>

#include "system/exception.h"

#include "inspiration/world.h"

#include "impl.h"
#include "wndresize.h"
#include "menulib.h"

using namespace pwn;

class MenuResize : public MenuEditorWindow {
	typedef MenuEditorWindow super;
protected:
	EditorImpl *ed;
public:
	gc::Pointer<Text> ok, cancel;
	
	gc::Pointer<TextField> width, height;
	gc::Pointer<TextField> tagwidth, tagheight;
	gc::Pointer<TextField> layers;
	
	MenuResize(EditorImpl *ed_);
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
};

class EditorWndResizeImpl {
public:
	gc::GlobalPointer<MenuResize> resizeWnd;

	EditorWndResizeImpl() : resizeWnd(gc::sys->createPtr<MenuResize>()) {}

	void init(EditorImpl *ed);
	void resize(resize_params_t *params);
};

MenuResize::MenuResize(EditorImpl *ed_)
:	super(vec2f(1,1),vec2f(25,9)),
	ok(createPtr<Text>()),
	cancel(createPtr<Text>()),
	width(createPtr<TextField>()),
	height(createPtr<TextField>()),
	tagwidth(createPtr<TextField>()),
	tagheight(createPtr<TextField>()),
	layers(createPtr<TextField>()),
	ed(ed_)
{
	visible = false;
	
	addChild(new Text(vec2f(1, 1), "Width:"));
	addChild(width = new TextField(vec2f(13, 1), 10));
	addChild(new Text(vec2f(1, 2), "Height:"));
	addChild(height = new TextField(vec2f(13, 2), 10));
	addChild(new Text(vec2f(1, 3), "Tag Width:"));
	addChild(tagwidth = new TextField(vec2f(13, 3), 10));
	addChild(new Text(vec2f(1, 4), "Tag Height:"));
	addChild(tagheight = new TextField(vec2f(13, 4), 10));
	addChild(new Text(vec2f(1, 5), "Layers:"));
	addChild(layers = new TextField(vec2f(13, 5), 10));

	addChild(ok = new Text(vec2f(7, 7), "OK"));
	addChild(cancel = new Text(vec2f(10, 7), "Cancel"));
}

eventResponse_t MenuResize::mouseEvent(MouseEvent e, const vec2f &v) {
	if (e.lpress) {
		if (selectedChild.ptr() == ok) {
			int iwidth = atoi(width->getText().c_str());
			int iheight = atoi(height->getText().c_str());
			int itagwidth = atoi(tagwidth->getText().c_str());
			int itagheight = atoi(tagheight->getText().c_str());
			int ilayers = atoi(layers->getText().c_str());
			ed->world->resizeMap(vec3i(iwidth, iheight, ilayers));
			ed->world->tags->setTagSize(vec2i(itagwidth, itagheight));
			visible = false;
		} else if (selectedChild.ptr() == cancel) {
			visible = false;
		}
	}
	
	super::mouseEvent(e,v);
	return RESPONSE_STOP;
}

void EditorWndResizeImpl::init(EditorImpl *ed) {
	resizeWnd = new MenuResize(ed);
	if (!resizeWnd) throw Exception() << "EditorWndResizeImpl::init failed";
	editor->rootmenu->addChild(resizeWnd);
}

//useful, consider moving up in the ladder
const char *ap(const char *_format, ...) {
	static char text[1024];
	va_list ap;
	va_start(ap, _format);
	vsnprintf(text, sizeof(text), _format, ap);
	va_end(ap);
	return text;
}

void EditorWndResizeImpl::resize(resize_params_t *params) {
	//upload info
	resizeWnd->width->setText(ap("%d", params->width));
	resizeWnd->height->setText(ap("%d", params->height));
	resizeWnd->tagwidth->setText(ap("%d", params->tag_tile_width));
	resizeWnd->tagheight->setText(ap("%d", params->tag_tile_height));
	resizeWnd->layers->setText(ap("%d", params->layers));
	//and show the window
	resizeWnd->setVisible(true);
}

EditorWndResize::EditorWndResize() : impl(new EditorWndResizeImpl()) {}
EditorWndResize::~EditorWndResize() { delete impl; }
void EditorWndResize::init(EditorImpl *ed) { impl->init(ed); }
void EditorWndResize::resize(resize_params_t *params) { impl->resize(params); }

