/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include "system/gl.h"
#include "blend.h"

// **************************
// BLEND FUNCTION COMPRESSION
// **************************

static unsigned int blendSrcMap[16] = {
	0,
	GL_ZERO,
	GL_ONE,
	GL_DST_COLOR,
	GL_ONE_MINUS_DST_COLOR,
	GL_SRC_ALPHA,
	GL_ONE_MINUS_SRC_ALPHA,
	GL_DST_ALPHA,
	GL_ONE_MINUS_DST_ALPHA,
	GL_SRC_ALPHA_SATURATE,

	GL_ZERO, GL_ZERO, GL_ZERO,GL_ZERO,GL_ZERO,GL_ZERO
};

static unsigned int blendDstMap[16] = {
	0,
	GL_ZERO,
	GL_ONE,
	GL_SRC_COLOR,
	GL_ONE_MINUS_SRC_COLOR,
	GL_SRC_ALPHA,
	GL_ONE_MINUS_SRC_ALPHA,
	GL_DST_ALPHA,
	GL_ONE_MINUS_DST_ALPHA,

	GL_ZERO,GL_ZERO,GL_ZERO,GL_ZERO,GL_ZERO,GL_ZERO,GL_ZERO
};

Blend CompressBlend(unsigned int sfactor, unsigned int dfactor)
{
	int i;
	Blend res;

	/*
	i would use bit fields (unsigned lo : 4; unsigned hi : 4;)
	but it turns out my compiler will auto round up to the nearest 4 byts
	and i like my Blend with only one byte
	*/
	for (i = 0; i < 16; i++) {
		if (blendSrcMap[i] == sfactor) {
			res = i;
			break;
		}
	}

	for (i = 0; i < 16; i++) {
		if (blendDstMap[i] == dfactor) {
			res = res | (i << 4);
			break;
		}
	}

	return res;
}

unsigned int BlendSFactor(Blend blend) {
	return blendSrcMap[blend&0x0F];
}

unsigned int BlendDFactor(Blend blend) {
	return blendDstMap[(blend & 0xF0) >> 4];
}

Blend getBlend(int method) {
	switch (method) {
	case BLEND_NONE:		return CompressBlend(GL_ONE, GL_ZERO);
	case BLEND_DECAL:		return CompressBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	case BLEND_ADD:			return CompressBlend(GL_ONE, GL_ONE);
	case BLEND_ADD_ALPHA:	return CompressBlend(GL_SRC_ALPHA, GL_ONE);
	case BLEND_DETAIL:		return CompressBlend(GL_DST_COLOR, GL_ZERO);
	}
	return CompressBlend(GL_ONE, GL_ZERO);
}