/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <fstream>

#include "filesys.h"

#ifdef WIN32
#include <direct.h>
#define getcwd _getcwd
#endif

using namespace std;

class FileSysImpl {
protected:
	string basedir;
public:
	bool sysInit();
	bool search(const string &filename, const vector<string> &vec, string &result);
	bool exists(const string& fileName);

};

bool FileSysImpl::sysInit() {
	//keep track of where we are
	char *_basedir = getcwd(0, 0);
	if (_basedir) {
		basedir = _basedir;
		basedir += "/";
		free(_basedir);
	}
	return true;
}

bool FileSysImpl::exists(const string& fileName) {
	fstream fin;
	fin.open(fileName.c_str(),ios::in);
	bool open = fin.is_open();
	fin.close();
	return open;
}
	
//translate a filename into a relative, searchable one
bool FileSysImpl::search(const string &filename, const vector<string> &vec, string &result) {
	//first try absolute!
	if (exists(filename)) {
		result = filename;
		return true;
	}
	//then do search paths:
	for (int i = 0; i < vec.size(); i++) {
		string searchFn = basedir + vec[i] + filename;
		if (exists(searchFn)) {
			result = searchFn;
			return true;
		}
	}
	return false;
}

FileSys::FileSys() : impl(new FileSysImpl()) {}
FileSys::~FileSys() { delete impl; }
bool FileSys::sysInit() { return impl->sysInit(); }
bool FileSys::search(const string &filename, const vector<string> &v, string &result) { return impl->search(filename, v, result); }
FileSys fileSys;
