/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
/*
code to still move over to the dll

  server-side:
entity physics - levels of simulation: none, static, AABB, (orientable BB?,) convex hull, (arbitrary hull?)
server net send ent & client stats
server net recv client input

  client-side
client net send client input
client net recv ent & client stats
client entity draw calls
client HUD draw calls

  what isn't handled anyways?
  rendering code
  animation framework
  map structure (some should be moved over, eh?)
  spawn object handling
  texture handling
  network connection / packet sending base code
  entity allocation code
  lighting code
  the level scripter . . . which isnt being used anyways
*/

#include <sstream>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "system/mouse.h"		//MouseHandler
#include "system/keyboard.h"	//Key

#include "base/base.h"

#include "inspiration/game/main.h"
#include "inspiration/game/spawn.h"
#include "inspiration/game/physics.h"
#include "inspiration/game/client.h"
#include "inspiration/game/program.h"

#include "inspiration/client.h"
#include "inspiration/world.h"
#include "inspiration/entity.h"
#include "inspiration/entalloc.h"
#include "inspiration/program.h"
#include "inspiration/menu.h"
#include "inspiration/animation.h"
#include "inspiration/main.h"
#include "inspiration/spawnobj.h"
#include "inspiration/strcasecmp.h"
#include "inspiration/lua.h"

using namespace std;

//client:

// *********************** UNIVERSAL USEFUL FUNCTIONS

void *Entity::operator new(size_t size) {
	return entAllocSys.newEnt(size);
}

void Entity::operator delete(void *_e) {
	Entity *e = (Entity *)_e;

	//unlink from all active entities
	for (int i = 0; i < entAllocSys.getCount(); i++) {
		Entity *f = entAllocSys.entForIndex(i);
		if (e == f) continue;	//no need to unlink
		f->unlink(e);
	}
	//(and from all serverside connection ents ... ?)

	entAllocSys.deleteEnt(_e);
}

//for engine delete calls -- since we dont want it messing with the ents on its own
void GameAPI::deleteEntity(entptr_t e) {
	Entity *ent = entAllocSys.entForPtr(e);
	if (!ent || !ent->active) return;
	
	lua_State *L = progSys.getState();
	if (!L) return;
	int top = lua_gettop(L);
	try {
		lua_pushentity(L, e);			//[ent]
		lua_getInstFunc(L, "delete");	//[ent:delete] [ent] 
		int ret = lua_pcall(L, 1, 0, 0);
		if (ret) throw Exception() << "couldn't run function.  got errors:\n" << lua_tostring(L,-1);
	} catch(Exception &t) {
		con << "failed to run entity:delete() for ent " << e.i << " : " << t << "\n";
		//in such a circumstance should we just remove it from the engine anyways?
	}
	lua_settop(L,top);
}

void Entity::think() {
	assert(active);
	lua_State *L = progSys.getState();
	int top = lua_gettop(L);
	lua_getglobal(L, "_ents");					// [_ents]
	assert (lua_istable(L,-1));
	
	entptr_t ep = entAllocSys.ptrForEnt(this);
	lua_pushinteger(L, ep.i );	// [_ents] [this._ptr]
	lua_rawget(L, -2);			// [_ents] [this]
	if (!lua_istable(L,-1)) {
//this condition will be gone by the time everything is ported into script
	} else {
		lua_getfield(L, -1, "think");					// [_ents] [this] [this:think]
		if (lua_isfunction(L, -1)) {
			lua_insert(L, lua_gettop(L)-1);				// [_ents] [this:think] [self]
			
			if (lua_pcall(L, 1, 0, 0)) {				// [_ents]
				con << "lua_pcall failed to call think:" << endl
					<< lua_tostring(L, -1) << endl
					<< lua_tostring(L, -2) << endl;
			}
		}
	}
	lua_settop(L,top);
}

Entity::Entity(SpawnObj *_spawnobj) {
	//ENTITY_CTOR_CONTENTS();	//this->__ctor();
	//begin contents:
	active = true;
	spawnobj = NULL;
	nextthink = -1;
	//end contents:
	spawnobj = _spawnobj;
	spawnobj->rw("pos") >> pos;
}

REGISTER_ENTITY(Entity)

void Renderable::ctor() {
	color.set(255,255,255,255);
	shader = 0;
	blend = BLEND_DECAL;
	scale = 1.f;
}

Renderable::Renderable(SpawnObj *s) : super(s) {
	ctor();

	string group;
	if (s->rw("group") >> group) {
		anim.setGroup(group.c_str());
	} else {
		anim.setGroup("null");	//or minus one? how do we turn off the sprite?
	}
	
	s->rw("angle") >> angle;
}

REGISTER_ENTITY(Renderable)


//NOTICE - ANYTHING that calls super::think (with super referring to the script think code)
//risks the entity dying (unless I add in a 'delete at the end of the frame' capability)
//AND THEREFORE needs to test whether the ent is still active or not ...
//GOOD THING this is almost completely phased out

REGISTER_ENTITY(Collidable)

class GameAPIImpl {
public:
	vector<string> clientKeys;
	vector<string> luaBaseIncludes;
	vector<string> luaPaths;
	vector<string> spritePaths;
	vector<string> texturePaths;
	vector<string> mapPaths;
};

entptr_t GameAPI::spawnPlayer(int conn, int local) {
	lua_State *L = progSys.getState();
	int top = lua_gettop(L);
	entptr_t ep = {-1};
	try {
		lua_getGameAPIfunction(L, "spawnPlayer");	//[spawnPlayer]
		lua_pushinteger(L, conn);
		lua_pushinteger(L, local);			//[spawnPlayer] [GameAPI] [conn] [local]
		int ret = lua_pcall(L, 3, 1, 0);
		if (ret) throw Exception() << "couldn't run function.  got errors:\n" << lua_tostring(L,-1);
		if (!lua_isnil(L,-1)) {
			if (lua_istable(L,-1)) {
				lua_rawgetfield(L, -1, "_ptr");
			}
			ep.i = (int)lua_tonumber(L, -1);
		}
	} catch(Exception &t) {
		con << "failed to run GameAPI:spawnPlayer: " << t << "\n";
		lua_printbacktrace(L);
	}
	lua_settop(L,top);
//	con << "GameAPI::spawnPlayer(" << conn << ") got entity " << ep.i << "\n";
	return ep;
}

// *********************** SPAWNOBJ / ENTITY INTERACTIONS

//issue: if someone calls 'delete' on an ent <= the current within a think func
//then our index will cause us to skip an ent . . .
//solution? 1) cycle backwards.  2) support a Entity *nextEnt(e) function
void GameAPI::entSystemUpdate() {

	//animation loop
	for (int i = entAllocSys.getCount()-1; i >= 0; i--) {
		//HERE'S AN ASSERTION ISSUE
		//if a think function or something did some removing ...
		//then we'll still be outta luck at asserting our entity is active
		//the ONLY safe way to do this is by dead-flagging things and removing them in one safe location
		//(or offloading all the work on Lua -- more on this later)
		Entity *e = entAllocSys.entForIndex(i);
		if (!e) continue;
		assert(e->active);

		Renderable *r;
		if (r = dynamic_cast<Renderable *>(e)) {
			//check for frame increment
			if (r->anim.frameNextTime < game_thistime) {
				//con.print("pre-increment-frame: %d\n", r->anim.getAbsFrame());
				r->anim.incFrame();
				//con.print("post-increment-frame:%d\n", r->anim.getAbsFrame());
			}
		}

		//think function
		if (!e->active) continue;
		if (e->nextthink < game_thistime) {
			e->think();
		}
	}
}

// *********************** system post mod init / shutdown ***********************

int getInputKeyCodeForName(const char *keyName) {
	assert(Key::name[numberof(Key::name)-1] == Key::nameEnd);
	for (int i = 0; i < Key::COUNT; i++) {
		if (!strcasecmp(keyName, Key::name[i])) {
			return i;
		}
	}
	//incentive to merge keyboard and mouse and all other buttons ...
	if (!strcasecmp(keyName, "mouseleft")) return ~MouseHandler::BUTTON_LEFT;
	if (!strcasecmp(keyName, "mouseright")) return ~MouseHandler::BUTTON_RIGHT;
	if (!strcasecmp(keyName, "mousemiddle")) return ~MouseHandler::BUTTON_MIDDLE;
	return 0;
}

static int panic(lua_State *L) {
	con.print("Lua Paniced!\n");
	con.print("%s\n", lua_tostring(L, -1));
	con.print("%s\n", lua_tostring(L, -2));
	return 0;
}

static void buildClientKeys(lua_State *L, int keyTable, short *clientKeyPtr) {
	assert(clientKeyPtr);
	int clientKeysTable = lua_gettop(L);
	lua_pushnil(L);
	while (lua_next(L, keyTable)) {
		const char *inkeyName = lua_tostring(L, -2);
		const char *inkeyValue = lua_tostring(L, -1);
		int clientKeyIndex = -1;
		bool found = false;
		for (int i = 0; i < game->clientKeys.size(); i++) {
			if (game->clientKeys[i] == inkeyName) {
				clientKeyIndex = i;
				found = true;
				break;
			}
		}
		if (!found) {
			if (game->clientKeys.size() >= MAX_INPUT_KEYS) {
				throw Exception() << "controls.lua had too many unique client keys";
			}
			clientKeyIndex = game->clientKeys.size();
			game->clientKeys.push_back(inkeyName);
		}
		short keyCode = getInputKeyCodeForName(inkeyValue);
		//convoluted names: keyboard keys vs key/value keys ...
		clientKeyPtr[clientKeyIndex] = keyCode;
		lua_pop(L,1);
	}
}

//called post entity routine allocation / default spawn key additions
// i think its safe to move the texture loading and key loading here,
// but only as long as you load them before calling "loadWorld" the first time
// AND (very important 'and') make sure, in the engine, that this is called before Editor_RefreshAll
//	(or else whatever spawn keys you add wont be reflected in the editor)
//
//so basically, for now, better just use this for the first map loading
bool GameAPI::gameInit() {
	//load our first world / start our initial scripts

//here are my thoughts for this one ...
//1) create a new lua state
//2) run the thing
//3) pull back from it certain globals...
//		initworld <= the filename for the first world to load
//		baseinclude <= an array of strings of base files to include
	lua_State *L = NULL;
	int err = 0;
	char init_world[256] = {'\0'};
	char init_menu[256] = {'\0'};
	
	game->luaBaseIncludes.resize(0);
	game->luaPaths.resize(0);
	game->spritePaths.resize(0);
	game->texturePaths.resize(0);
	game->mapPaths.resize(0);
	game->clientKeys.resize(0);
	mapTypes.resize(0);
	
	game->spritePaths.push_back(string("./"));	//default search path	- always need at least one
	game->texturePaths.push_back(string("./"));
	game->mapPaths.push_back(string("./"));
	
	try {
		if (!(L = lua_open())) throw Exception() << "failed to open new lua state";
		lua_atpanic(L, panic);
		luaL_openlibs(L);
		luaStateAddFuncs(L, frameworkGlueFuncs);	//rest of the story - framework calls
		if (err = luaL_loadfile(L, "init.lua")) {
			throw Exception() << "failed to load init.  got errors: " << err << "\n" << lua_tostring(L, -1) << "\n" << lua_tostring(L,-2);
		}

		//runbase:
		//begin rip from lua.exe: - i guess this is for the callback for pcall?
		int base = lua_gettop(L);  /* function index */
		lua_pushliteral(L, "_TRACEBACK");
		lua_rawget(L, LUA_GLOBALSINDEX);  /* get traceback function */
		lua_insert(L, base);  /* put it under chunk and args */
		if (err = lua_pcall(L, 0, 0, 0)) {
			throw Exception() << "lua_pcall got error " << err << "\n" << lua_tostring(L,-1) << "\n" << lua_tostring(L, -2);
		}
		lua_remove(L, base);  /* remove traceback function */
		//end rip
		
		//now extract the globals
		lua_getglobal(L, "initworld");
		const char *l_initworld = lua_tostring(L, -1);
		if (!l_initworld) throw Exception() << "didn't find the 'initworld' lua variable!";
		strncpy(init_world, l_initworld, sizeof(init_world));

		lua_getglobal(L, "initmenu");
		const char *l_initmenu = lua_tostring(L, -1);
		if (!l_initmenu) {
			con << "didn't find the 'initmenu' lua variable!  bypassing..." << endl;
			memset(init_menu, 0, sizeof(init_menu));
		} else {
			strncpy(init_menu, l_initmenu, sizeof(init_menu));
		}
		
		LuaTable global(L, LUA_GLOBALSINDEX);
		global.rw("baseinclude") >> game->luaBaseIncludes;
		global.rw("luapaths") >> game->luaPaths;
		global.rw("spritepaths") >> game->spritePaths;
		global.rw("texpaths") >> game->texturePaths;
		global.rw("mappaths") >> game->mapPaths;
		global.rw("maptypes") >> mapTypes;
		
		//i should make a separate file for stuff the game will read and write
		//config.lua
		// and in it goes: player keys, video stuff, which mod to run, etc
		
		lua_getglobal(L, "menukeys");
		cout << "reading client menu keys..." << endl;
		if (lua_istable(L, -1)) {
			buildClientKeys(L, lua_gettop(L), clientSys.getMenuKeyPtr());					
		}
		cout << "done reading client menu keys..." << endl;
		lua_pop(L,1);
		
		//while we're here, map all the client keys out
		//todo - put this into its own section
		//make a player profile class or something
		lua_getglobal(L, "playerkeys");
		if (lua_istable(L,-1)) {
			int t = lua_gettop(L);
			// table is in the stack at index 't'
			int clientIndex = 0;
			lua_pushnil(L);  // first key
			while (lua_next(L, t)) {
				// uses 'key' (at index -2) and 'value' (at index -1)
				if (lua_istable(L, -1)) {
					if (clientIndex >= MAX_LOCAL_CLIENTS) throw Exception() << "controls.lua had too many players";
					
					/*
					ClientKeySet
						vector<string> command;		<- shared by all
						per client:
							vector<string> key;		<- indexes match up with commands ...
							
						this means, to have separate menu keys, you will have to share command vector
						which is cool
						up, down, left, right
						primary / ok / fire / attack
						secondary / cancel / alt ...
						think of some standard ...
					*/
					
					//this operates on the game->clientKeys vector
					buildClientKeys(L, lua_gettop(L), clientSys.getKeyPtr(clientIndex));					
				}
					
				// removes 'value'; keeps 'key' for next iteration
				lua_pop(L, 1);
				clientIndex++;
			}
		}

	} catch (Exception &t) {
		con << "GameAPI::gameInit got error:\n" << t << "\n";
	}
	if (L) lua_close(L);
	
	vector<vector<string>*> pathVectors;
	pathVectors.push_back(&game->spritePaths);
	pathVectors.push_back(&game->texturePaths);
	pathVectors.push_back(&game->luaPaths);
	pathVectors.push_back(&game->mapPaths);
	
	//make sure the paths have a trailing / on all our path vectors
	for (int j = 0; j < pathVectors.size(); j++) {
		vector<string> *pathVector = pathVectors[j];
		for (vector<string>::iterator i = pathVector->begin(); i != pathVector->end(); i++) {
			string &s = *i;
			if (!s.length()) s = s + ".";	//empty string means '.' (or remove it from the array)
			if (s[s.length()-1] != '/') s = s + "/";	//make sure there's a trailing slash
		}
	}	
	
	//depends upon spawn type names
	//load world file
	//TODO - make these game api variables along with the paths
	//have their respective systems reference them from the gameapi
	worldSys.setInitWorld(init_world);
	mainMenuSys.setInitMenuFile(init_menu);

	return true;
}

//game-scope, need to expose this to the menu...
// maps from key name to index in the array
int PF_getClientKeyMap(lua_State *L) {
	lua_newtable(L);
	int table = lua_gettop(L);
	for (int i = 0; i < game->clientKeys.size(); i++) {
		lua_pushstring(L, game->clientKeys[i].c_str());
		lua_pushinteger(L, i);
		lua_settable(L, table);
	}
	return 1;
}

// ************************** session init / shutdown ***************************

bool GameAPI::sessionInit() {
	//TODO - why run main?  why not just run the root level code?

	//init all level variables here!

	//make sure our client spawn type is set to the player start:
//	strcpy(client_spawn_typename, "player_start");

	//and run the main function!
	//run the "main" method
	//and wait for it to terminate (to allow the camera, fade, etc, to all set up)
//	con.print("trying to run the main function...\n");
	lua_State *L = progSys.getState();
	lua_getglobal(L, "main");
	if (!lua_isfunction(L, -1)) {
//		con.print("failed to find main script function!\n");
		lua_pop(L,1);
	} else {
		int ret = lua_pcall(L, 0, 0, 0);
		if (ret) {
			con.print("failed to run main. got errors:\n");
			con.print("%s\n", lua_tostring(L, -1));
			con.print("%s\n", lua_tostring(L, -2));
		}
	}
	
	return true;
}

void GameAPI::sessionKill() {
}

void GameAPI::sysKill() {
	//this is all called in the engine - get rid of it here?
	spawnTypeSys.clear();
}

GameAPI::GameAPI() : impl(new GameAPIImpl()) {}
GameAPI::~GameAPI() { delete impl; }

const vector<string> &GameAPI::getSpritePaths() { return game->spritePaths; }
const vector<string> &GameAPI::getTexturePaths() { return game->texturePaths; }
const vector<string> &GameAPI::getLuaBaseIncludes() { return game->luaBaseIncludes; }
const vector<string> &GameAPI::getLuaPaths() { return game->luaPaths; }
const vector<string> &GameAPI::getMapPaths() { return game->mapPaths; }

const vector<std::string> &GameAPI::getClientKeys() { return game->clientKeys; }

GameAPI game;
