/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include "inspiration/game/spawn.h"
#include "inspiration/spawnobj.h"
#include "inspiration/world.h"
#include "inspiration/entalloc.h"
#include "inspiration/animation.h"

using namespace std;

//pass it the current tag and the last recorded tag
//this function tests them, and, upon change, refreshes all objects,
//put this here so it has access to the spawn_entry_table
void spawnObjSystemUpdateTags(int nextTag, int lastTag) {

	World *world = worldSys.getWorld();
	assert(world);

	//int nextTag = mapGetTag(floor_ftoi(_x), floor_ftoi(_y));

	//respawning - clear the entities
//	if (this->dead) nextTag = -1;

	if (nextTag != lastTag) {

		//kill all old and spawn all new entities
		for (int i = 0; i < spawnObjSys.count(); i++) {
			SpawnObj *s = spawnObjSys.get(i);
			if (!s) continue;

			vec2f sv;
			s->rw("pos") >> sv;

			box2f bbox = s->getTypeBBox();
			box2i ibbox;
			ibbox.min.set( floor_ftoi(sv.x + bbox.min.x), floor_ftoi(sv.y + bbox.min.y) );
			ibbox.max.set( floor_ftoi(sv.x + bbox.max.x), floor_ftoi(sv.y + bbox.max.y) );

			//get the tags of the corners
			short tag1 = world->tags->maploc(ibbox.min.x, ibbox.min.y);
			short tag2 = world->tags->maploc(ibbox.min.x, ibbox.max.y);
			short tag3 = world->tags->maploc(ibbox.max.x, ibbox.max.y);
			short tag4 = world->tags->maploc(ibbox.max.x, ibbox.min.y);

			bool remove = tag1 == lastTag || tag2 == lastTag || tag3 == lastTag || tag4 == lastTag;
			bool create = tag1 == nextTag || tag2 == nextTag || tag3 == nextTag || tag4 == nextTag;
			
			//only process changes in state
			if (remove == create) continue;

			if (remove) {	//equals last tag - remove it
				entptr_t ep = s->ent;
				if (ep.i != -1) {	//ep is a block index, between 0 and BLOCK_COUNT
					Entity *e = entAllocSys.entForPtr(ep);
					if (e && e->active) {
						delete e;
					}
				}
			}
			
			//if we are creating a new entity
			if (create) {
				//and if it is not there already
				//FIXME - still more bugs with multiplayer...
				//ERROR HERE - what happens when we have . . . 
				//one room's spawnobj's point to a set of entities
				//the user leaves the room
				//the associated ents are removed
				//new ents are recreated in the old ent's places
				//the user goes back to the first room
				// . . . all the spawnObjGetEnt's point to active entities!
				
				//here's a very thoughtless solution:
				if (s->ent.i != -1) {
					delete entAllocSys.entForPtr(s->ent);
				}
				
//				if (s->ent.i == -1)
				{
					//then try to create it
					if (!s->createEnt()) {
						string classname = "::unknown::";
						s->rw("classname") << classname;
						con << "spawnObjSystemUpdateTags(): unable to spawn an entity for spawnobj " << i << " of type " << classname << "\n";
					}
				}
			}
		}

		//swap it in with the new one
//		lastTag = nextTag;
//		this->tagtype = MapTagKey_GetType(this->tag);
	}
}

SpawnObj *spawnObjFindNext(const char *spawntype, SpawnObj *lastptr) {
	int count = spawnObjSys.count();
	if (!count) return NULL;
	SpawnObj *sret = NULL;
	int last = spawnObjSys.getIndex(lastptr);
	int start = (last + count + 1) % count;
	int si = start;
	for (;;) {
		SpawnObj *s = spawnObjSys.get(si);
		if (!s) break;	//this should only happen when si = start = -1
		
		string classname;
		if (s->rw("classname") >> classname && classname.length()) {
			if (classname == spawntype) {
				sret = s;
				break;
			}
		}
		si = (si + 1) % count;
		if (si == start) break;
	}
	return sret;
}

Entity *spawnCopy(SpawnObj *s) {
	Entity *e = NULL;
	entptr_t last_ep = s->ent;
	if (s->createEnt()) {
		e = entAllocSys.entForPtr(s->ent);
	}
	s->ent = last_ep;
	return e;
}

//factory:
EntRegLink *EntRegLink::last = NULL;

Entity *EntRegLink::createByName(const char *name) {
	for (EntRegLink *l = last; l; l = l->prev) {
		if (!strcmp(l->name, name)) return l->create();
	}
	return NULL;
}
