/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <assert.h>
#include "system/exception.h"
#include "game/client.h"
#include "nettables.h"
#include "game/nettables.h"
#include "program.h"	//lua_printbacktrace

using namespace std;

/*
OH CRAP
to get this all working I need to complete my cl/sv model for Lua
even with my predefined structure
--
to do this
do the netSendTo() stuff on serverside
and a netRecvFrom() stuff clientside -- 
-- but see here, thats the challenge ...
-- we've always had this challenge ... how to minimize our protocol while supporting a range of types and what not...
-- here's my thoughts on it:
with every class() subclass of Entity we have it register a Net Entity Class signature #
then (so long as cl/sv are running the same scripts ... (pull shirt collar)) they will match up
(for fancy testing send the alignments as a text-to-# file or something)
-- Then the server will send ent class #'s followed by net table info
-- the client will interpret them, create a metaclass thing for accessing net table stuff
-- and then call the associated class' 'render' function (renamed to 'netRecvFrom') with 'self' as the net table stats
...
we'll have a basic implementation for Renderable so if the user really doesn't care then they won't have to
...
though it seems we should have an override for single-player (no-network) models ...
like maybe separate 'clientUpdate' and 'clientUpdateRemote' for each entity - called depending on whether we're running loopback or remote
(seems most appealing)

so we'll have - in the script -

0.5) upon client connect /slash/ level init (when all script is done building) we'll send a map from entity class names to net table #s
	from the server to the client.  this way the client will know what to make of what later on (translation key)

1) we'll need separate enums for the classes (to tell the client what method to run)
	and separate enums for the net tables (to tell the client what kind of table to use)
	ONLY PASS THE CLASS to the client - from there (assuming their script lines up) they can determine the assoc. net table

1) classes specify (w/inheritence) their net tables.  by default we'll just have one.

	-- #if local connections --

1.999) just run the 'clientUpdate' method of the entity

	-- #elif remote connections --

2) a netTable object meta-linked to the net send stuff (no more key/val string passing)

2.5) server send we fill the net table (scan across all registered keys)

3) server delta compresses and sends to the client, who delta uncompresses

4) client unravels, finds what class to call depending on the ent class enum, and calls its 'clientUpdateRemote' with 
	an object meta index linked to the net table stuffs



some issues on the topic ...
the hope with net tables is that we can have dynamically sized network code per-entity
this would be cool for an extensible engine
 but here are the issues
with variable size, well, you have to keep track of memory per entity
	and you have to keep track of a prev frame and a current frame (for delta compression sake)
	this is all still do-able
	here's how so:
in the script code when defining entity classes:

class('NetTable')

function netTable(classObj, params)
	-- the first {} is what is returned (i.e. 'create a new table')
	params = union({}, classObj.super.netTable, params)
	... do the engine stuff ... (setting up the net table and what not)
	... from here on out every Entity allocated will reference its lua class' associated net table to see if theres one to make as well
	classObj.netTable = ... the net table object (much like a SpawnObj instance)
end
-- then attach its __index and __newindex to the engine, just like SpawnObj
-- keep track of internal pointers via the _ptr, as we're all familiar with

-- this creates the nettable for the entity
-- the function takes care of reading from super's nettable
-- and stores the table inside of classObj.netTable
netTable(Entity, {
	active = NETTABLE.BOOL,
	pos = NETTABLE.VEC2,
})

-- this creates the nettable for our new class
-- the nettable function will know to search all superclasses for all nettable members
-- (this search can be kept track of engine-side or calculated upon netTableCreate execution)
class('Whatever', Entity)
SpawnTypeForClass('spawnname', 'groupname', bbox, 'spawndesc', Whatever)	-- this will set up the init bbox and group, as well as link the spawntype to the engine
netTableCreate(Whatever, {
	dog = NETTABLE.INT	
})


hey how about I put a max size on the net tables?
otherwise we're gonna have lots of allocation / unallocation every frame ...
aw screw it, it goes on anyways right?
besides we'll only have freeing and creating as often as we have entities entering and leaving the scene
not too often

[NETCMD_UPDATE_ENTITY]		// doubles as a create and an update (???)
[entity index]
[active]
[classID]
[delta compressed net table data]
=> if active is false then free any existing assoc nettable data and bail
=> if we get a change in classID then unalloc/realloc the nettable data (copy it over too for further deltas)
=> then delta-decompress it all

nettables storage:
class ClientEnt {
	//always the first two entries in the net table:
	bool active;

	unsigned short entClassID;		// used to reference to the net table descriptor.

	//rest of the story. i mean data.
	void *ptr;			// to the net table data: [active] [pos] etc ...
};


so we got net tables working
and we have lua extensibility working
but here's the problem

[ent server update] ->
 [send ent info to client] ->
  [fill server net table structures] ->
   [point the net table system to the net table associated with the current entity's class]
   [call lua "populate the net tables" function] ->
    [lua call net table update for each variable we want to send] ->
	 [net table store the passed value in the associated net table]
  [calc delta info on server tables]
  [send delta info to clients]
[ent client update] ->
 [client rebuilds ent structs] ->

*/

//these store the connective info between strings&values passed from scripts
//to raw data locations to be sent across the net connection

//our prototype table will be ClientEntity ...

NetTypeBool netTypeBool;
NetTypeChar netTypeChar("char");
NetTypeUChar netTypeUChar("uchar");
NetTypeShort netTypeShort("short");
NetTypeUShort netTypeUShort("ushort");
NetTypeInt netTypeInt("int");
NetTypeUInt netTypeUInt("uint");
NetTypeFloat netTypeFloat("float");
NetTypeDouble netTypeDouble("double");
NetTypeVec2f netTypeVec2f("vec2f");
NetTypeVec2s netTypeVec2s("vec2s");
NetTypeVec4s netTypeVec4s("vec4s");
NetTypeColor netTypeColor;
NetTypeAngle netTypeAngle;

static map<string,NetType*> netTypes;


NetTable::~NetTable() {
	for (map<string,NetTableEntry*>::iterator i = entries.begin(); i != entries.end(); i++) {
		delete i->second;
	}
	entries.clear();
}

//assert its a new'd entry
void NetTable::addEntry(const char *name, const NetType &type, bool engineUpdate) {
	map<string, NetTableEntry*>::iterator i = entries.find(name);
	if (i != entries.end()) {
		con << " -- DANGER -- NetTable::addEntry got a repeat entry for name " << name << "\n";
		return;
	}
	
	//todo here
	//if the name is on our engine-bound list
	//then instead of reading or writing to lua
	//read/write to the engine var
/*
here's what we need to do:
automate it all...

but lua needs callback functions - one for each variable ...
 ... unless i pass both the key and the value back through the engine ...
 ... which would be on par with generating the meta tables engine-side
 ... which is appealing
 
that means create, for each engine-bound class, a map from names to locations (and types)
then in our meta callbacks, look up the name and return the value
and then in our nettable key registration
	hunt through all class and parent classes , through all their registered names, to see if there's a shortcut
	if its there then use it.
*/
	entries[name] = new NetTableEntry(type,size,engineUpdate);
	size += type.getSize();
	
	//until i get dynamic-sized client entities
	assert (size <= sizeof(ClientEntity));
}

//explicit location
void NetTable::addEntry(const char *name, const NetType &type, size_t ofs, bool engineUpdate) {
	map<string, NetTableEntry*>::iterator i = entries.find(name);
	if (i != entries.end()) {
		con << " -- DANGER -- NetTable::addEntry got a repeat entry for name " << name << "\n";
		return;
	}
	
	entries[name] = new NetTableEntry(type,ofs,engineUpdate);
	ofs += type.getSize();	//move to the end of the entry
	//get the max of it and our current size
	if (size < ofs) size = ofs;

	//so this whole setup of pushing the size back as we explicitly state offsets
	//what that means is, when adding entries,
	//ALWAYS add the fixed offset entries first
}

static string getAllTypes() {
	string s = "";
	for (map<string,NetType*>::iterator i = netTypes.begin(); i != netTypes.end(); i++) {
		s = s + i->first + string("\n");
	}
	return s;
}

class NetTablesSysImpl {
public:
	NetTableInstance *currentInstance;
	vector<NetTable *> tables;
	vector<EntClass *> entClasses;
	
	NetTablesSysImpl() : currentInstance(NULL) {}
	NetTable *getTable(int tableIndex);
	EntClass *getEntClass(int entClassIndex);
	bool sysInit();
	bool sessionInit();
	void sessionKill();
};


NetTableInstance::NetTableInstance(int tableIndex_, void *dataPtr_, int entIndex_)
:	tableIndex(tableIndex_),
	dataPtr(dataPtr_),
	entIndex(entIndex_),
	dirtyFlag(false)
{
	//not multithread safe
	//if you want to make it mt-safe then put the dirty flag in the NetTableInstance class
	//then, when calling your lua callbacks that a NetTableInstance lifecycle encloses,
	//pass a reference to the NetTableInstance as a lightuserdata
	//then have the PF_netTable callback accept that lightuserdata and write to its dirty bit when something goes down
	assert(!netTablesSys->currentInstance);
	netTablesSys->currentInstance = this;
}

NetTableInstance::~NetTableInstance() {
	assert(netTablesSys->currentInstance);
	netTablesSys->currentInstance = NULL;
}


//here's my hack for letting the rest of the world get to the currently selected net table's data
void *NetTableInstance::rawAccess(const char *key) {
	try {
		NetTable *table = netTablesSys->getTable(tableIndex);
		if (!table) throw Exception() << "got a bad table index: " << tableIndex;
		map<string,NetTableEntry*>::iterator i = table->entries.find(key);
		if (i == table->entries.end()) return NULL;	//use null to tell the outside world we found no key

		NetTableEntry *entry = i->second;
		char *fieldPtr = (char*)dataPtr + entry->ofs;
		return fieldPtr;
	} catch (Exception &t) {
		con << "NetTableInstance.rawAccess: " << t << "\n";
	}
	return NULL;
}

NetTable *NetTableInstance::getCurrentTable() {
	return netTablesSys->getTable(tableIndex);
}

void NetTableInstance::deltaWrite(lua_State *L, char *deltamask) {
	int top = lua_gettop(L);
	NetTable *table = netTablesSys->getTable(tableIndex);
	if (!table) {
		con << "got a bad table index: " << tableIndex << endl;
		return;
	}
	for (map<string,NetTableEntry*>::iterator i = table->entries.begin(); i != table->entries.end(); ++i) {
		const string &key = i->first;
		NetTableEntry *entry = i->second;
		char *deltaptr = (char *)deltamask + entry->ofs;
		char *curptr = (char *)dataPtr + entry->ofs;
		char diff = 0;
		for (int i = 0; !diff && i < entry->type.getSize(); i++) {
			diff = deltaptr[i];
		}
		if (diff) {
			//con << "client->server writing table entry " << key << endl;
			entry->type.writeToLua(L, curptr);
			lua_setfield(L, top, key.c_str());
		}
	}
	lua_settop(L, top);
}


NetTable *NetTablesSysImpl::getTable(int tableIndex) {
	if (tableIndex < 0 || tableIndex >= tables.size()) return NULL;
	return tables[tableIndex];
}

EntClass *NetTablesSysImpl::getEntClass(int entClassIndex) {
	if (entClassIndex < 0 || entClassIndex >= entClasses.size()) return NULL;
	return entClasses[entClassIndex];
}

bool NetTablesSysImpl::sysInit() {	
	//init the net types
	//maybe they should be declared in this class instead of static ... ?
#define ADDNETTYPE(x)	netTypes[netType##x.name] = &netType##x;
	ADDNETTYPE(Bool)
	ADDNETTYPE(Char)
	ADDNETTYPE(UChar)
	ADDNETTYPE(Short)
	ADDNETTYPE(UShort)
	ADDNETTYPE(Int)
	ADDNETTYPE(UInt)
	ADDNETTYPE(Float)
	ADDNETTYPE(Double)
	ADDNETTYPE(Vec2f)
	ADDNETTYPE(Vec2s)
	ADDNETTYPE(Vec4s)
	ADDNETTYPE(Color)
	ADDNETTYPE(Angle)
	return true;
}

bool NetTablesSysImpl::sessionInit() {
	return true;
}

void NetTablesSysImpl::sessionKill() {
	VECTOR_DELETE_ALL(entClasses);
	VECTOR_DELETE_ALL(tables);
}

NetTablesSystem::NetTablesSystem() : impl(new NetTablesSysImpl()) {}
NetTablesSystem::~NetTablesSystem() { delete impl; }
bool NetTablesSystem::sysInit() { return impl->sysInit(); }
bool NetTablesSystem::sessionInit() { return impl->sessionInit(); }
void NetTablesSystem::sessionKill() { impl->sessionKill(); }
EntClass *NetTablesSystem::getEntClass(int index) { return impl->getEntClass(index); }
NetTableInstance *NetTablesSystem::getCurrentInstance() { return impl->currentInstance; }
NetTablesSystem netTablesSys;

//1- get the parent of the allocated class
//2- return the new index
int PF_registerEntClass(lua_State *L) {
	try {
		//what we'll need:
		//the parent class ID
		//and the parent's net table ID which we can lookup this side of things
		
		lua_getfield(L, 1, "super");		//[class.super]
		int parentID = -1;
		if (lua_isnil(L, -1)) {	//then this is Entity, or it better be at least
		} else {
			lua_getfield(L, -1, "_entClassID");	//[class.super] [class.super._entClassID]
			parentID = (int)lua_tonumber(L, -1);
			lua_pop(L,1);
		}
		lua_pop(L,1);
		//make sure we're in bounds
		if (parentID < -1 || parentID >= (int)netTablesSys->entClasses.size()) throw Exception() << "got a bad parent ID: " << parentID;
		
		//get the associated net ID.  it is -1 for Entity and the parent's net ID otherwise
		//you'd think this would keep things -1, but netTableCreate overrides this so all is well
		int parentNetID = -1;
		if (parentID != -1) parentNetID = netTablesSys->entClasses[parentID]->netTableID;
		
		//get the current class ID.
		int id = netTablesSys->entClasses.size();
		
		NetTable *entTable = NULL;
#ifdef HANDLE_ENTITY_META_IN_ENGINE
		//see if there's an engine-bound table waiting for us in game/program.cpp
		//DIRTY DIRTY BOY!
		lua_getfield(L, 1, "classname");
		const char *classname = lua_tostring(L, -1);
		if (classname) {
			if (!strcmp(classname, "Entity")) {
				entTable = &keyTable_Entity;
			} else if (!strcmp(classname, "Renderable")) {
				entTable = &keyTable_Renderable;
			} else if (!strcmp(classname, "Collidable")) {
				entTable = &keyTable_Collidable;
			}
		}
		lua_pop(L,1);
#endif
		
		//set it in the entClasses vector
		netTablesSys->entClasses.push_back(new EntClass(parentID, parentNetID, entTable));

		//set it in lua
		lua_pushinteger(L, id);
		lua_setfield(L, 1, "_entClassID");
	} catch (Exception &t) {
		con << "PF_registerEntClass failed: " << t << "\n";
	}
	return 0;
}


int PF_netTableCreate(lua_State *L) {
	//todo : cleanup appropriately on thrown exception
	try {
		if (!lua_istable(L, 1)) throw Exception() << "expected param 1 to be a table";
/*
	for making a new net table
	how should inheritence be handled?
	
	well we're gonna have a table from the entity class to the net table, right?
	or maybe we should have a net table index per-class instead?
	then when inheritence looks up what net table to use, it'll find the right one.
*/
		NetTable *table = new NetTable();
		if (!table) throw Exception() << "failed to alloc net table!";
		
		int top = lua_gettop(L);
		
		//get the current class name
		lua_getfield(L, 1, "classname");
		table->name = lua_tostring(L, -1);
		lua_pop(L, 1);
		
		//get the current class ID
		lua_getfield(L, 1, "_entClassID");
		int entClassID = (int)lua_tonumber(L, -1);
		lua_pop(L, 1);
		if (entClassID < 0 || entClassID >= netTablesSys->entClasses.size()) throw Exception() << "got an invalid ent class id " << entClassID;

		//now get the next parent-most netTableID
		int netTableParent = -1;
		int parentClassID = netTablesSys->entClasses[entClassID]->parentClassID;
		if (parentClassID != -1) netTableParent = netTablesSys->entClasses[parentClassID]->netTableID;		
		
		//for this tables's addEntry we need to cycle up the parent chain and see if any of them have 
		//a key matching
		
		//if there is no net table parent then add all the default values to it
		if (netTableParent == -1) {
			//these entries should be on all net tables
			//but definitely not associated with the entity class
			//since not all entities are sent over the wire
			//these must be added first
			table->addEntry("active", netTypeBool, OFFSETOF(ClientEntity, active), true);	//looks like bools take 2 bytes in mactel
			table->addEntry("entClassID", netTypeShort, OFFSETOF(ClientEntity, entClassID), true);
			table->addEntry("pos", netTypeVec2f, OFFSETOF(ClientEntity, pos), true);
			//...
			//they're no longer in all net tables
			//now their reading from entities is hardwired in entUpdateNetInfo
			//of course, they (and the subsequent hardwired entries) still need their room in the net tables...
			//they just need another parameter to distinctify that they'll be directly written to from the engine callbacks rather than via lua field reads...
		} else {
			//otherwise add all the parents' entries
			//dont forget to add copies so deleting the tables wont crash
			NetTable *parent = netTablesSys->getTable(netTableParent);
			if (!parent) throw Exception() << "tried to get an oob net table " << netTableParent;
			for (map<string,NetTableEntry*>::iterator i = parent->entries.begin(); i != parent->entries.end(); i++) {
				table->addEntry(i->first.c_str(), i->second->type, i->second->ofs, i->second->engineUpdate);
			}
		}

		
		//renderable is special -- has engine bound fields
		if (table->name == "Renderable") {
			//these are linked to the params updated in PF_animUpdateNetInfo
			//of course PF_colUpdateNetInfo doesn't do anything since no fields are updated by Collidable
			//...and PF_entUpdateNetInfo updates the bare bones basics for any entity: active, entClassID, and pos
			//	(which are added in the parent-less netTableCreate function)
			const char *params[][2] = {
				{"absFrame", "int"},
				{"angle", "angle"},
				{"left", "bool"},
				{"scale", "vec2f"},
				{"blend", "uchar"},
				{"color", "color"},
			};
			
			for (int i = 0; i < numberof(params); i++) {
				const char *key = params[i][0];
				const char *type_name = params[i][1];
			
				map<string,NetType*>::iterator j = netTypes.find(type_name);
				if (j == netTypes.end()) throw Exception() << "got an unknown type " << type_name << " for key " << key << ". possible types are:\n" << getAllTypes();
				
				bool engineUpdate = true;
				table->addEntry(key, *j->second, engineUpdate);
			}
		}

	//iterate through the second parameter
		//if we got a table to iterate ...
		
		if (lua_istable(L, 2)) {
			lua_pushnil(L);
			int top = lua_gettop(L);
			while (lua_next(L, 2)) {
				if (!lua_isstring(L, -2)) throw Exception() << "somehow got a key that wasnt a string";
				const char *key = lua_tostring(L,-2);
				
//here - if our name is pre-registered as a hardcoded name then 
// we should add some sort of special entry or something ...
//you know i'm thinking more and more that it'd be faster to just lua everything
//i think the aforenoticed slowdown wasn't due to my pseudo attempt to lua-ize some things
//but rather due to the remaining C bindings needing to be called as often as they are
				
				if (!lua_isstring(L, -1)) throw Exception() << "found a value that wasnt a string";
				const char *type_name = lua_tostring(L,-1);
				
				map<string,NetType*>::iterator i = netTypes.find(type_name);
				if (i == netTypes.end()) throw Exception() << "got an unknown type " << type_name << " for key " << key << ". possible types are:\n" << getAllTypes();
				
				/*
				rather than allow this to be set here
				I think it'd be best of the Renderable nettable was created and bound at first
				and same with same with the Collidable
				*/
				bool engineUpdate = false;

				table->addEntry(key, *netTypes[type_name], engineUpdate);
				lua_pop(L,1);
			}
		}
/*
		con << "new table\n";
		for (map<string,NetTableEntry*>::iterator i = table->entries.begin(); i != table->entries.end(); i++) {
			con << " " << i->first << " ofs: " << i->second->ofs << "\n";
		}
		con << "table size is " << table->getSize() << "\n";
*/
		int netTableID = netTablesSys->tables.size();	//get its index
		netTablesSys->tables.push_back(table);	//add it to the list

		//and update the netTableID in our entClasses[]
		netTablesSys->entClasses[entClassID]->netTableID = netTableID;
	} catch (Exception &t) {
		con << "PF_netTableCreate got an error: " << t << "\n";
		lua_printbacktrace(L);
	}
	return 0;
}

int PF_netTable(lua_State *L) {
	try {
		int argc = lua_gettop(L);
		if (argc < 1) throw Exception() << "usage: (value = ) PF_netTable('param', value)";

		const char *key = lua_tostring(L, 1);
		if (!key) throw Exception() << "expected a valid string as a key";
		
		//TODO - to make this mt-safe make a parameter out of it
		NetTableInstance *inst = netTablesSys->currentInstance;
		if (!inst) throw Exception() << "got a null inst";

		NetTable *table = inst->getCurrentTable();
		
		//special hack for the read only _ptr key 
		//for getting the ent's index
		//that information isn't stored in the net object, and isn't exposed to the clientside any other way
		if (!strcmp(key, "_ptr")) {
			lua_pushinteger(L, inst->getEntIndex());
			return 1;
		}
		
		map<string,NetTableEntry*>::iterator i = table->entries.find(key);
		if (i == table->entries.end()) {
			throw Exception() <<  "in table " << table->name << " couldn't find key " << key;
		}
		NetTableEntry *entry = i->second;
		char *fieldPtr = (char*)inst->getDataPtr() + entry->ofs;

		//TODO - size check?
		//NOTICE HOW IT WONT DO BOTH READ AND WRITE - JUST ONE OR THE OTHER
		if (argc == 1) return entry->type.writeToLua(L, fieldPtr);
		inst->setDirtyFlag();
		entry->type.readFromLua(L, 2, fieldPtr);
	} catch (Exception &t) {
		con << "PF_netTable failed: " << t << "\n";
		//should we session kill?
		lua_printbacktrace(L);
	}
	return 0;
}
