require 'base.scripts.ents'
require 'base.scripts.defs'

-- you know, consider separating collidable and walkable for the Door and Chest's sake

-- spin-off of the ents.lua Collidable class
-- but this one isn't engine-bound so it doesnt do traces during the think
-- instead it does manual map tests on integer tilesets
-- and interpolates between them on the off season
--
-- superclass for anything that moves.
--
entclass('LoloBase', Renderable)

LoloBase.bbox = box2:new(.4)
LoloBase.destPos = vec2:new()	-- because ctors access setPos, accessing our override, before we have a chance to set up our var

function LoloBase:ctor(so)
	LoloBase.super.ctor(self,so)

	self.startPos = vec2:new(self:getPos())
	self.srcPos = vec2:new(self:getPos())
	self.destPos = vec2:new(self:getPos())
	
	self.moveCmd = nil	-- this says where to go - one of the SIDE enums

	-- oh yeah,
	-- since we're undermining the collidable class
	-- i need to do its bbox stuff here now
	self.bbox = box2:new()

	-- override by spawn type
	if self.class.bbox then self.bbox:set(self.class.bbox) end
	
	-- override by inline
	if so then
		if so.bbox then self.bbox:set(so.bbox) end
	end
end

-- used externally for any direct position setters
-- not for interpolation - that needs direct variable access
function LoloBase:setPos(pos)
	LoloBase.super.setPos(self,pos)
	self.destPos:set(pos)
end

-- by default all collidable objects will have their 'cant walk thru me' flag set
--	what should pushable objects do with this flag?
LoloBase.isBlocking = true

-- likeewise, by default, all collidable objects will have their 'cant push over me' flag set
-- separate of above because the two will vary
LoloBase.isBlockingPushers = true

-- where we hit
-- returns true if we stop at the edge
function LoloBase:hitEdge(where) return true end

-- static function me?
function LoloBase:cannotPassThru(maptype)
	return maptype ~= MAPTYPE.Empty
end

-- where we hit
-- types = types of corners of our movement (everything moves by half tiles)
-- returns true if we have to stop
function LoloBase:hitWorld(where, typeUL, typeUR, typeLL, typeLR)
	return
		self:cannotPassThru(typeUL) or
		self:cannotPassThru(typeUR) or
		self:cannotPassThru(typeLL) or
		self:cannotPassThru(typeLR)
end

-- responses to hitObject
-- STOP means we stop altogether (and test nothing more)
-- MOVE_THRU means we move through the object (and dont test it via startPush)
-- TEST_OBJECT means we move, but only if the object's startPush lets us
LoloBase.HIT_RESPONSE = enum({'STOP', 'MOVE_THRU', 'TEST_OBJECT'})

-- what is being pushed
-- pushDest is where self is going to
-- side is what dir
function LoloBase:hitObject(what, pushDest, side)
	return self.HIT_RESPONSE.TEST_OBJECT
end

-- who is pushing
-- pushDest where who will stand
-- side we're getting pushed towards
-- returns true if we block our pusher (keep them from moving)
function LoloBase:startPush(who, pushDest, side) return self.isBlocking end

-- who is pushing
-- pushDest where who will ended up
function LoloBase:endPush(who, pushDest) end

-- this one's messy.  dependant upon shot which is dependant upon this.
-- called by shot:hitObject - consider a call graph graph reduction between the two
function LoloBase:hitByShot(shot, shotPos, shotDir)
	-- if we hit a blocking object then kill the shot
	if self.isBlocking then
		shot:delete()
		return self.HIT_RESPONSE.STOP
	end
	return self.HIT_RESPONSE.MOVE_THRU
end

-- here's another class variable that goes backwards in the hierarchy
-- its used by the chest
-- do nothing by default
-- door is gonna open
-- enemies and eggs will die
function LoloBase:onChestOpen() end
