-- first create the basics of our clientview
require 'base.scripts.clientview'
require 'base.scripts.hud'
require 'base.scripts.vec'
require 'base.scripts.util'	--timedRepeat, enum
require 'topview.scripts.clientdb'

equipFields = {'LHand','RHand','Head','Body','Relic1','Relic2','Esper1','Esper2'}

-- i could subclass that and use its subclass ...
-- that would be more proper ...
-- more open for extensibility if we wanted something to reference both this and base
entclass('TopviewClientView', ClientView)

-- need defaults for these nettable members or else the server will complain that it can't access them
TopviewClientView.gauge1 = 0
TopviewClientView.gauge2 = 0
TopviewClientView.gauge3 = 0
TopviewClientView.gauge4 = 0
TopviewClientView.health1 = 0
TopviewClientView.health2 = 0
TopviewClientView.health3 = 0
TopviewClientView.health4 = 0

function TopviewClientView:ctor(so)
	TopviewClientView.super.ctor(self,so,true)
	
	self.clientUid = tonumber(so.clientUid)
	self.clientInfo = clients[self.clientUid]
	if self.clientInfo then
		assert(not self.clientInfo.clientView, "got two client infos for one client")
		self.clientInfo.clientView = self
	end
	
	-- how about making partyChar into partyChars, mapping from local indices to map ents?
	-- then spawn more for battles -- whoever isnt there?
	self.partyChars = table.new()
	if not so.noParty and not so.target then
		self:spawnPartyChars{1}
		self:followTarget(self.partyChars[1])
	end
end

function TopviewClientView:spawnPartyChars(args)
	if not self.clientInfo then return end
	local dir = args.dir or SIDE.DOWN
	local pos
	if args.pos then
		pos = vec2:new(args.pos)
	else
		pos = vec2:new(self:getPos())
	end
	for k,v in ipairs(args) do
		-- accepts maps from naturals => index
		--  or maps from index => true
		--	or maps from index => args
		--		where args is a table of {pos=...}
		local i
		local charPos 
		local charDir = dir
		if not type(v) == 'number' then
			assert(type(k) == 'number')
			i = k
			if type(v) == 'table' then
				charPos = v.pos
				charDir = v.dir
			end
		else
			i = v
		end
		if not charPos then
			if dir == SIDE.UP then
				charPos = pos + vec2:new(i-3,1)
			elseif dir == SIDE.LEFT then
				charPos = pos + vec2:new(1,i-3)
			elseif dir == SIDE.RIGHT then
				charPos = pos + vec2:new(-1,i-3)
			else	--if dir == SIDE.DOWN then
				charPos = pos + vec2:new(i-3,-1)
			end
		end
		local charIndex = self.clientInfo.chars[i]
		if not self.partyChars[i] then
			self.partyChars[i] = Char:new{
				pos=charPos;
				dir=charDir;
				charIndex=charIndex;
				onMsg='CharFightOnMsg';		-- TODO - organizeme globals...
				onDmg='CharReviveOnDmg';	-- come back to life after a second of damage
			}
		end
	end
end

function TopviewClientView:removePartyChars(args)
	if not self.clientInfo then return end
	for _,i in ipairs(args) do
		if self.partyChars[i] then
			self.partyChars[i]:delete()
			self.partyChars[i] = nil
		end
	end
end

function TopviewClientView:followTarget(target)
	if self.target then self.target.clientview = nil end	--unbind
	TopviewClientView.super.followTarget(self,target)
	if self.target then self.target.clientview = self end	--rebind
end

-- clientside call

function TopviewClientView.clientUpdate(classObj)
	TopviewClientView.super.clientUpdate(classObj)

	-- isn't really needed since we don't render anything anyways
	-- TopviewClientView.super.clientUpdate()
	if net.screenFlash[4] then
		screenFlash(net.screenFlash)
	end
end

TopviewClientView:createNetTable{
	screenFlash = 'color';
	-- for the battle system
	-- this could be bypassed if it was an entity member
	-- and if the client had access to the entity array
	--		(reflection anyone?)
	gauge1 = 'float';
	gauge2 = 'float';
	gauge3 = 'float';
	gauge4 = 'float';
	health1 = 'float';
	health2 = 'float';
	health3 = 'float';
	health4 = 'float';
}


createClientStartClass(TopviewClientView, 'marker', box2:new(.4))



