require 'base.scripts.util'	--timedRepeat
require 'base.scripts.misc'	-- Light

animLoad('well', {sprites = {{file = 'well.png'; bbox = {-1,-1,1,1}}}})
Spawnable(entclass('Well',Collidable,{animGroup='well',bbox=box2:new(1)}))

entclass('CampfireFire', Particle, {
	ctor = function(self,pos)
		local shade = math.random()
		local scalar = math.random()
		CampfireFire.super.ctor(self,{
			pos = pos + vec2:new(math.random()-.5, math.random()-.75)*.5;
			vel = vec2:new(0, -1);
			radius = scalar * .4;
			color = vec4:new(1,shade,shade*0.5,1);
			life = 2 * (1 - scalar) * (1 - scalar) + math.random() * 0.3 + 0.3;
		})
		self.baseScale = vec2:new(self:getScale())
	end;
	
	think = function(self)
		CampfireFire.super.think(self)
		if not self:isActive() then return end
		local frac = self.life / self.srclife	-- from 1 to 0
		self:setScale(self.baseScale * frac)
		self:setColorW(1)	-- because usually it fades by frac
	end;
})

animLoad('firewood', 'firewood.png')
entclass('Campfire', Renderable, {
	animGroup = 'firewood';
	bbox = box2:new(1);
	lightRadius = 10;
	lightColor = vec4:new(1);
	makeFlame = true;

	ctor = function(self,so)
		Campfire.super.ctor(self,so)
		
		-- start a light off at our position
		self.light = Light:new{pos=so.pos, radius=self.lightRadius}
		self.light.flicker = true
	end;
	
	think = function(self)
		Campfire.super.think(self)
		-- make some particles
		if self.makeFlame then
			CampfireFire:new(self:getPos())
		end
	end;
	
	startUp = function(self, dt)
		self.light.enabled = true
		self.makeFlame = true
		self:fadeLight(self.lightColor, dt)
	end;
	
	goOut = function(self, dt)
		-- fork a thread to do the following:
		-- 1) gradually dim light
		-- 2) slowly fade out smoke (hey we don't have any smoke particles going on . . .)

		self.makeFlame = false
		self:fadeLight(vec4:new(0), dt)

		-- once the fire is out, disable it
		self.light.enabled = false
	end;
	
	fadeLight = function(self,dest,dt)
		local src = vec4:new(self.light:getColor())
		timedRepeat(function(t)
			local f = t / dt
			self.light:setColor(vec4:new():lerp(src, dest, f))
		end, dt)
		self.light:setColor(dest)
	end;
})
Spawnable(Campfire)
