require 'topview.scripts.itemdb'
require 'topview.scripts.clientview'
require 'topview.scripts.char'

-- as long as stores are only created in global namespace
--they will match between clients and servers
-- (works the same as the callbacks of the ClientSide class)

class('Store')
Store.stores = {}	-- static singleton

function Store:ctor(args)
	local id = args.id or #Store.stores+1
	self.storeID = id
	self.items = table.new()
	for _,v in ipairs(args) do
		local itemIndex
		if type(v) == 'string' then
			itemIndex = itemIndexFromName[v]
		else
			itemIndex = tonumber(v)
		end

		if not itemIndex then
			print('store ',self.storeID,' requested unknown item ',v)
		else
			self.items:insert(itemIndex)
		end
	end
	Store.stores[self.storeID] = self
end

-- TODO - storedb?
Store:new{
	'Mithril Knife';
	'Buckler';
	'Leather Armor';
	'Leather Hat';
}

entclass('Merchant', Char)
Spawnable(Merchant)

function Merchant:ctor(so)
	Merchant.super.ctor(self, so)
	self.store = Store.stores[tonumber(so.store) or 1]	-- todo - name-associate and name-lookup here
end

function Merchant:doInteract(other)
	if not self.store then return end
	if other.clientview then
		other.clientview:shop(self.store.storeID)
	end
end

-- this, like dialog(), must be run on a separate thread
function TopviewClientView:shop(storeID)
	self:createPromptingMenu{
		class='StoreMenu';
		storeID=storeID;
	}
end

function TopviewClientView:serverStorePurchase(storeID, itemIndex)
	storeID = tonumber(storeID)
	itemIndex = tonumber(itemIndex)
	local store = assert(Store.stores[storeID])
	local item = assert(items[itemIndex])
	if self.clientInfo.gold >= item.price then
		self.clientInfo.gold = self.clientInfo.gold - item.price
		local itemLocalIndex = table.find(self.clientInfo.items, itemIndex, function(v, index) return v.item == index end)
		if itemLocalIndex then
			self.clientInfo.items[itemLocalIndex].amount = self.clientInfo.items[itemLocalIndex].amount + 1
		else
			table.insert(self.clientInfo.items, {item=itemIndex, amount=1})
		end
		print("purchased ", item.name)
		-- TODO send a message back to refresh the menu
	else
		print("you can't afford this")
	end
end

	-- clientside
	
class('StoreMenu', Menu)
StoreMenu.pos = {1,1}
StoreMenu.size = { Menu.sysSize[1]-2, 10}

function StoreMenu:onCancel()
	self:delete()
end

function StoreMenu:ctor(args)
	StoreMenu.super.ctor(self, args)
	assert(self.storeID)
	local store = assert(Store.stores[self.storeID])

	-- list the items.  scrollbar somehow?
	-- should I expose constructors for the menu lib controls?
	local scrollArea = Menu:new{
		parent=self;
--		class="ScrollArea";
		pos={1,1};
		size=self.size-{2,2};
	}
	-- scroll area?  this will be tricky...
	local scrollContainer = Menu:new{
		parent=scrollArea;
		pos={1,1};
		size=scrollArea.size-{2,2};
		hasPrompts=true;
	}
	local width = scrollContainer.size[1]
	
	local storeMenu = self
	for i,itemIndex in ipairs(store.items) do
		local item = items[itemIndex]
		local itemRow = Menu:new{
			pos={0,(i-1)*1.5};
			size={width*.6, 1.5};
			parent=scrollContainer;
			storeID=self.storeID;
			itemIndex=itemIndex;
			onOk = function(self)
				-- highlight all children on mouseover
				--	add highlight on enter / remove highlight on exit
				--	this requires callbacks for focus() / blur()
				entClientSendMessage(self.client.entID, 'serverStorePurchase', self.storeID, self.itemIndex)
			end;
			onCancel = function(self)
				storeMenu:delete()
			end;
			isPrompt=true;
		}
		local itemmenu = Menu:new{
			pos={0,0};
			text=item.name;
			size={width * .3, 1.5};
			parent=itemRow;
		}
		local pricemenu = Menu:new{
			pos={width * .3, 0};
			text=tostring(item.price);
			size={width * .3, 1.5};
			parent=itemRow;
		}
	end
end

